"""
Tests of the ``clifford.g*`` modules and friends.

Only used for modules which are not yet tested elsewhere """

import importlib
import pytest
import sys


all_algebras = {
    'g2', 'g2c', 'g3', 'g3_1', 'g3c', 'g4', 'gac', 'pga', 'sta'
}

# algrebras which are tested elsewhere
tested_algebras = {'g3c'}


@pytest.mark.parametrize('alg_name', all_algebras - tested_algebras)
def test_import(alg_name):
    """ Test that this file is importable, assuming nothing else is testing it """
    assert 'clifford.{}'.format(alg_name) not in sys.modules, "already tested elsewhere"

    importlib.import_module('clifford.{}'.format(alg_name))


@pytest.mark.parametrize('alg_name', tested_algebras)
def test_already_imported(alg_name):
    """ Test that some other test has already imported this module to test it """
    assert 'clifford.{}'.format(alg_name) in sys.modules



import numbers
import math
from typing import List, Set, Tuple

import numpy as np

import clifford as cf
from . import (
    general_exp,
    grades_present,
    compute_reordering_sign_and_canonical_form,
)


class MultiVector(object):
    """An element of the algebra

    Parameters
    -------------
    layout: instance of :class:`clifford.Layout`
        the layout of the algebra

    value : sequence of length ``layout.gaDims``
        the coefficients of the base blades

    Notes
    ------
    The following operators are overloaded:

    * ``A * B`` : geometric product
    * ``A ^ B`` : outer product
    * ``A | B`` : inner product
    * ``A << B`` : left contraction
    * ``~M`` : reversion
    * ``M(N)`` : grade or subspace projection
    * ``M[N]`` : blade projection
    """

    def __init__(self, layout, value=None, string=None, *, dtype: np.dtype = np.float64) -> None:
        """Constructor."""

        self.layout = layout
        self.__array_priority__ = 100

        if value is None:
            if string is None:
                self.value = np.zeros((self.layout.gaDims,), dtype=dtype)
            else:
                self.value = layout.parse_multivector(string).value
        else:
            self.value = np.array(value)
            if self.value.shape != (self.layout.gaDims,):
                raise ValueError(
                    "value must be a sequence of length %s" %
                    self.layout.gaDims)

    def __array__(self) -> 'cf.MVArray':
        # we are a scalar, and the only appropriate dtype is an object array
        return cf.MVArray([self])

    def _checkOther(self, other, coerce=True) -> Tuple['MultiVector', bool]:
        """Ensure that the other argument has the same Layout or coerce value if
        necessary/requested.

        _checkOther(other, coerce=True) --> newOther, isMultiVector
        """
        if isinstance(other, numbers.Number):
            if coerce:
                # numeric scalar
                newOther = self._newMV(dtype=np.result_type(other))
                newOther[()] = other
                return newOther, True
            else:
                return other, False

        elif isinstance(other, MultiVector):
            if other.layout != self.layout:
                raise ValueError(
                    "cannot operate on MultiVectors with different Layouts")
            else:
                return other, True
        else:
            return other, False

    def _newMV(self, newValue=None, *, dtype: np.dtype = None) -> 'MultiVector':
        """Returns a new MultiVector (or derived class instance).
        """
        if newValue is None and dtype is None:
            raise TypeError("Must specify either a type or value")

        return self.__class__(self.layout, newValue, dtype=dtype)

    # numeric special methods
    # binary

    def exp(self) -> 'MultiVector':
        return general_exp(self)

    def vee(self, other) -> 'MultiVector':
        """
        The vee product aka. the meet... To be optimised still
        """
        return self.layout.MultiVector(value=self.layout.vee_func(self.value, other.value))

    def __and__(self, other) -> 'MultiVector':
        """
        The vee product aka. the meet... To be optimised still
        """
        return self.vee(other)

    def __mul__(self, other) -> 'MultiVector':
        """Geometric product :math:`MN` """

        other, mv = self._checkOther(other, coerce=False)

        if mv:
            newValue = self.layout.gmt_func(self.value, other.value)
        else:
            if isinstance(other, np.ndarray):
                obj = self.__array__()
                return obj*other

            newValue = other * self.value

        return self._newMV(newValue)

    def __rmul__(self, other) -> 'MultiVector':
        """Right-hand geometric product, :math:`NM`"""

        other, mv = self._checkOther(other, coerce=False)

        if mv:
            newValue = self.layout.gmt_func(other.value, self.value)
        else:
            if isinstance(other, np.ndarray):
                obj = self.__array__()
                return other*obj
            newValue = other*self.value

        return self._newMV(newValue)

    def __xor__(self, other) -> 'MultiVector':
        r""" Outer product, :math:`M \wedge N` """

        other, mv = self._checkOther(other, coerce=False)

        if mv:
            newValue = self.layout.omt_func(self.value, other.value)
        else:
            if isinstance(other, np.ndarray):
                obj = self.__array__()
                return obj^other
            newValue = other*self.value

        return self._newMV(newValue)

    def __rxor__(self, other) -> 'MultiVector':
        r"""Right-hand outer product, :math:`N \wedge M` """

        other, mv = self._checkOther(other, coerce=False)

        if mv:
            newValue = self.layout.omt_func(other.value, self.value)
        else:
            if isinstance(other, np.ndarray):
                obj = self.__array__()
                return other^obj
            newValue = other * self.value

        return self._newMV(newValue)

    def __or__(self, other) -> 'MultiVector':
        r""" Inner product, :math:`M \cdot N` """

        other, mv = self._checkOther(other)

        if mv:
            newValue = self.layout.imt_func(self.value, other.value)
        else:
            if isinstance(other, np.ndarray):
                obj = self.__array__()
                return obj|other
            # l * M = M * l = 0 for scalar l
            return self._newMV(dtype=np.result_type(self.value.dtype, other))

        return self._newMV(newValue)

    __ror__ = __or__

    def __add__(self, other) -> 'MultiVector':
        """Addition

        M + N
        """

        other, mv = self._checkOther(other)
        if not mv:
            if isinstance(other, np.ndarray):
                obj = self.__array__()
                return obj + other
        newValue = self.value + other.value

        return self._newMV(newValue)

    __radd__ = __add__

    def __sub__(self, other) -> 'MultiVector':
        """Subtraction

        M - N
        """

        other, mv = self._checkOther(other)
        if not mv:
            if isinstance(other, np.ndarray):
                obj = self.__array__()
                return obj - other
        newValue = self.value - other.value

        return self._newMV(newValue)

    def __rsub__(self, other) -> 'MultiVector':
        """Right-hand subtraction

        N - M
        """

        other, mv = self._checkOther(other)
        if not mv:
            if isinstance(other, np.ndarray):
                obj = self.__array__()
                return other - obj
        newValue = other.value - self.value

        return self._newMV(newValue)

    def right_complement(self) -> 'MultiVector':
        return self.layout.MultiVector(value=self.layout.right_complement_func(self.value))

    def __truediv__(self, other) -> 'MultiVector':
        """Division, :math:`M N^{-1}`"""

        other, mv = self._checkOther(other, coerce=False)

        if mv:
            return self * other.inv()
        else:
            if isinstance(other, np.ndarray):
                obj = self.__array__()
                return obj/other
            newValue = self.value / other
            return self._newMV(newValue)

    def __rtruediv__(self, other) -> 'MultiVector':
        """Right-hand division, :math:`N M^{-1}`"""

        other, mv = self._checkOther(other)
        if isinstance(other, np.ndarray):
            obj = self.__array__()
            return other / obj

        return other * self.inv()

    def __pow__(self, other) -> 'MultiVector':
        """Exponentiation of a multivector by an integer, :math:`M^{n}` """

        if not isinstance(other, (int, float)):
            raise ValueError("exponent must be a Python int or float")

        if abs(round(other) - other) > cf._eps:
            raise ValueError("exponent must have no fractional part")

        other = int(round(other))

        if other == 0:
            return 1

        newMV = self._newMV(np.array(self.value))  # copy

        for i in range(1, other):
            newMV = newMV * self

        return newMV

    def __rpow__(self, other) -> 'MultiVector':
        """Exponentiation of a real by a multivector, :math:`r^{M}`"""

        # Let math.log() check that other is a Python number, not something
        # else.

        # pow(x, y) == exp(y * log(x))
        newMV = general_exp(math.log(other) * self)

        return newMV

    def __lshift__(self, other) -> 'MultiVector':
        """
        The ``<<`` operator is the left contraction
        """
        return self.lc(other)

    # unary

    def __neg__(self) -> 'MultiVector':
        """Negation, :math:`-M`"""

        newValue = -self.value

        return self._newMV(newValue)

    def as_array(self) -> np.ndarray:
        return self.value

    def __pos__(self) -> 'MultiVector':
        """Positive (just a copy), :math:`+M` """

        newValue = self.value + 0  # copy

        return self._newMV(newValue)

    def mag2(self) -> numbers.Number:
        """Magnitude (modulus) squared, :math:`{|M|}^2`

        Note in mixed signature spaces this may be negative
        """
        mv_val = self.layout.gmt_func(self.layout.adjoint_func(self.value), self.value)
        return mv_val[0]

    def __abs__(self) -> numbers.Number:
        """Magnitude (modulus), :math::`|M|`

        This is ``sqrt(abs(~M*M))``.

        The abs inside the sqrt is need for spaces of mixed signature
        """

        return np.sqrt(abs(self.mag2()))

    def adjoint(self) -> 'MultiVector':
        r"""Adjoint / reversion, :math:`\tilde M`

        Aliased as ``~M`` to reflect :math:`\tilde M`, one of several
        conflicting notations.

        Note that ``~(N * M) == ~M * ~N``.
        """
        # The multivector created by reversing all multiplications
        return self._newMV(self.layout.adjoint_func(self.value))

    __invert__ = adjoint

    # builtin
    def __int__(self) -> int:
        """Coerce to an integer iff scalar.
        """

        return int(self.__float__())

    def __float__(self) -> float:
        """"Coerce to a float iff scalar.
        """

        if self.isScalar():
            return float(self[()])
        else:
            raise ValueError("non-scalar coefficients are non-zero")

    # sequence special methods
    def __len__(self) -> int:
        """Returns length of value array.
        """

        return self.layout.gaDims

    def __getitem__(self, key) -> numbers.Number:
        """If key is a blade tuple (e.g. (0, 1) or (1, 3)), or a blade,
        (e.g. e12),  then return the (real) value of that blade's coefficient.
        Otherwise, treat key as an index into the list of coefficients.

        value = M[blade]
        value = M[index]
        """
        if isinstance(key, MultiVector):
            return self.value[int(np.where(key.value)[0][0])]
        elif key in self.layout.bladeTupMap.keys():
            return self.value[self.layout.bladeTupMap[key]]
        elif isinstance(key, tuple):
            sign, blade = compute_reordering_sign_and_canonical_form(key, self.layout.sig,
                                                                     self.layout.firstIdx)
            return sign*self.value[self.layout.bladeTupMap[blade]]
        return self.value[key]

    def __setitem__(self, key, value: numbers.Number) -> None:
        """If key is a blade tuple (e.g. (0, 1) or (1, 3)), then set
        the (real) value of that blade's coeficient.
        Otherwise treat key as an index into the list of coefficients.

        M[blade] = value
        M[index] = value
        """
        if key in self.layout.bladeTupMap.keys():
            self.value[self.layout.bladeTupMap[key]] = value
        elif isinstance(key, tuple):
            sign, blade = compute_reordering_sign_and_canonical_form(key, self.layout.sig,
                                                                     self.layout.firstIdx)
            self.value[self.layout.bladeTupMap[blade]] = sign*value
        else:
            self.value[key] = value

    def __delitem__(self, key) -> None:
        """Set the selected coefficient to 0.

        del M[blade]
        del M[index]
        """

        if key in self.layout.bladeTupMap.keys():
            self.value[self.layout.bladeTupMap[key]] = 0
        elif isinstance(key, tuple):
            sign, blade = compute_reordering_sign_and_canonical_form(key, self.layout.sig,
                                                                     self.layout.firstIdx)
            self.value[self.layout.bladeTupMap[blade]] = 0
        else:
            self.value[key] = 0

    # grade projection
    def __call__(self, other, *others) -> 'MultiVector':
        """Return a new multi-vector projected onto a grade OR a MV


        M(grade[s]) --> <M>
                        grade
        OR

        M(other) --> other.project(M)

        Examples
        --------
        >>>M(0)
        >>>M(0, 2)
        """
        if isinstance(other, MultiVector):
            return other.project(self)
        else:
            # we are making a grade projection
            grade = other

        if len(others) != 0:
            return sum([self.__call__(k) for k in (other,)+others])

        if grade not in self.layout.gradeList:
            raise ValueError("algebra does not have grade %s" % grade)

        if not np.issubdtype(type(grade), np.integer):
            raise ValueError("grade must be an integer")

        mask = np.equal(grade, self.layout.gradeList)

        newValue = np.multiply(mask, self.value)

        return self._newMV(newValue)

    # fundamental special methods
    def __str__(self) -> str:
        """Return pretty-printed representation.
        """

        s = ''
        p = cf._print_precision

        for grade, name, coeff in zip(self.layout.gradeList, self.layout.names, self.value):
            # if we have nothing yet, don't use + and - as operators but
            # use - as an unary prefix if necessary
            if s:
                seps = (' + ', ' - ')
            else:
                seps = ('', '-')

            if abs(coeff) >= cf._eps:
                if coeff > 0:
                    sep = seps[0]
                    abs_coeff = round(coeff, p)
                else:
                    sep = seps[1]
                    abs_coeff = -round(coeff, p)

                if grade == 0:
                    # scalar
                    s = '%s%s%s' % (s, sep, abs_coeff)
                else:
                    # not a scalar
                    s = '%s%s(%s^%s)' % (s, sep, abs_coeff, name)
        if s:
            # non-zero
            return s
        else:
            # return scalar 0
            return '0'

    def __repr__(self) -> str:
        """Return eval-able representation if global _pretty is false.
        Otherwise, return str(self).
        """

        if cf._pretty:
            return self.__str__()

        s = "MultiVector(%s, value=%s)" % (
            repr(self.layout), list(self.value))
        return s

    def __bool__(self) -> bool:
        """Instance is nonzero iff at least one of the coefficients
        is nonzero.
        """

        nonzeroes = np.absolute(self.value) > cf._eps

        if nonzeroes.any():
            return True
        else:
            return False

    def __eq__(self, other) -> bool:
        other, mv = self._checkOther(other)
        if not mv:
            return NotImplemented

        if (np.absolute(self.value - other.value) < cf._eps).all():
            # equal within epsilon
            return True
        else:
            return False

    def __ne__(self, other) -> bool:
        ret = self.__eq__(other)
        if ret is NotImplemented:
            return ret
        return not ret

    def clean(self, eps=None) -> 'MultiVector':
        """Sets coefficients whose absolute value is < eps to exactly 0.

        eps defaults to the current value of the global cf._eps.
        """

        if eps is None:
            eps = cf._eps

        mask = np.absolute(self.value) > eps

        # note element-wise multiplication
        self.value = mask * self.value

        return self

    def round(self, eps=None) -> 'MultiVector':
        """Rounds all coefficients according to Python's rounding rules.

        eps defaults to the current value of the global cf._eps.
        """

        if eps is None:
            eps = cf._eps

        self.value = np.around(self.value, eps)

        return self

    # Geometric Algebraic functions
    def lc(self, other) -> 'MultiVector':
        r"""The left-contraction of two multivectors, :math:`M\rfloor N`"""

        other, mv = self._checkOther(other, coerce=True)

        newValue = self.layout.lcmt_func(self.value, other.value)

        return self._newMV(newValue)

    @property
    def pseudoScalar(self) -> 'MultiVector':
        "Returns a MultiVector that is the pseudoscalar of this space."
        return self.layout.pseudoScalar

    I = pseudoScalar

    def invPS(self) -> 'MultiVector':
        "Returns the inverse of the pseudoscalar of the algebra."

        ps = self.pseudoScalar

        return ps.inv()

    def isScalar(self) -> bool:
        """Returns true iff self is a scalar.
        """

        indices = list(range(self.layout.gaDims))
        indices.remove(self.layout.gradeList.index(0))

        for i in indices:
            if abs(self.value[i]) < cf._eps:
                continue
            else:
                return False

        return True

    def isBlade(self) -> bool:
        """Returns true if multivector is a blade.
        """
        if len(self.grades()) != 1:
            return False

        return self.isVersor()

    def isVersor(self) -> bool:
        """Returns true if multivector is a versor.
        From Leo Dorsts GA for computer science section 21.5, definition from 7.6.4
        """
        Vhat = self.gradeInvol()
        Vrev = ~self
        Vinv = Vrev/(self*Vrev)[0]

        # Test if the versor inverse (~V)/(V * ~V) is truly the inverse of the
        # multivector V
        if grades_present(Vhat*Vinv, 0.000001) != {0}:
            return False
        if not np.sum(np.abs((Vhat*Vinv).value - (Vinv*Vhat).value)) < 0.0001:
            return False

        # applying a versor (and hence an invertible blade) to a vector should
        # not change the grade
        if not all(
            grades_present(Vhat*e*Vrev, 0.000001) == {1}
            for e in cf.basis_vectors(self.layout).values()
        ):
            return False

        return True

    def grades(self) -> Set[int]:
        """Return the grades contained in the multivector.

        .. versionchanged:: 1.1.0
            Now returns a set instead of a list
        """
        return grades_present(self, cf._eps)

    @property
    def blades_list(self) -> List['MultiVector']:
        '''
        ordered list of blades present in this MV
        '''
        blades_list = self.layout.blades_list
        value = self.value

        b = [value[0]] + [value[k]*blades_list[k] for k in range(1, len(self))]
        return [k for k in b if k != 0]

    def normal(self) -> 'MultiVector':
        r"""Return the (mostly) normalized multivector.

        The _mostly_ comes from the fact that some multivectors have a
        negative squared-magnitude.  So, without introducing formally
        imaginary numbers, we can only fix the normalized multivector's
        magnitude to +-1.

        :math:`\frac{M}{|M|}` up to a sign
        """

        return self / np.sqrt(abs(self.mag2()))

    def leftLaInv(self) -> 'MultiVector':
        """Return left-inverse using a computational linear algebra method
        proposed by Christian Perwass.
        """
        return self._newMV(self.layout.inv_func(self.value))

    def normalInv(self) -> 'MultiVector':
        r"""The inverse of itself if :math:`M \tilde M = |M|^2`.

        ..math::

            M^{-1} = \tilde M / (M \tilde M)
        """

        Madjoint = ~self
        MadjointM = (Madjoint * self)

        if MadjointM.isScalar() and abs(MadjointM[()]) > cf._eps:
            # inverse exists
            return Madjoint / MadjointM[()]
        else:
            raise ValueError("no inverse exists for this multivector")

    def inv(self) -> 'MultiVector':
        if (self*~self).isScalar():
            it = self.normalInv()
        else:
            it = self.leftLaInv()
        return it

    leftInv = leftLaInv
    rightInv = leftLaInv

    def dual(self, I=None) -> 'MultiVector':
        r"""The dual of the multivector against the given subspace I, :math:`\tilde M = MI^{-1}`

        I defaults to the pseudoscalar.
        """
        if I is None:
            return self.layout.MultiVector(value=self.layout.dual_func(self.value))
        else:
            Iinv = I.inv()

        return self * Iinv

    def commutator(self, other) -> 'MultiVector':
        r"""The commutator product of two multivectors.

        :math:`[M, N] = M \cross N = (MN + NM)/2`
        """

        return ((self * other) - (other * self)) / 2

    x = commutator

    def anticommutator(self, other) -> 'MultiVector':
        """The anti-commutator product of two multivectors, :math:`(MN + NM)/2` """

        return ((self * other) + (other * self)) / 2

    def gradeInvol(self) -> 'MultiVector':
        r"""The grade involution of the multivector.

        .. math::
            M^* = \sum_{i=0}^{\text{dims}}
                  {(-1)^i \left<M\right>_i}
        """

        signs = np.power(-1, self.layout.gradeList)

        newValue = signs * self.value

        return self._newMV(newValue)

    @property
    def even(self) -> 'MultiVector':
        '''
        Even part of this multivector

        defined as
        ``M + M.gradInvol()``
        '''
        return .5*(self + self.gradeInvol())

    @property
    def odd(self) -> 'MultiVector':
        '''
        Odd part of this mulitvector

        defined as
        ``M +- M.gradInvol()``
        '''
        return .5*(self - self.gradeInvol())

    def conjugate(self) -> 'MultiVector':
        """The Clifford conjugate (reversion and grade involution).

        :math:`M^*` = ``(~M).gradeInvol()``
        """
        return (~self).gradeInvol()

    # Subspace operations
    def project(self, other) -> 'MultiVector':
        r"""Projects the multivector onto the subspace represented by this blade.

        :math:`P_A(M) = (M \rfloor A) A^{-1}`
        """

        other, mv = self._checkOther(other, coerce=True)

        if not self.isBlade():
            raise ValueError("self is not a blade")

        return other.lc(self) * self.inv()

    def factorise(self) -> Tuple[List['MultiVector'], numbers.Number]:
        """
        Factorises a blade into basis vectors and an overall scale.

        Uses Leo Dorsts algorithm from 21.6 of GA for Computer Science
        """
        if not self.isBlade():
            raise ValueError("self is not a blade")
        scale = abs(self)
        max_index = np.argmax(np.abs(self.value))
        B_max_factors = self.layout.bladeTupList[max_index]

        factors = []

        B_c = self/scale
        for ind in B_max_factors[1:]:
            ei = self.layout.blades_list[ind]
            fi = (ei.lc(B_c)*(~B_c*(1/(B_c*~B_c)[0]))).normal()
            factors.append(fi)
            B_c = B_c * ~fi * (1 / (fi * ~fi)[0])
        factors.append(B_c.normal())
        factors.reverse()
        return factors, scale

    def basis(self) -> List['MultiVector']:
        """Finds a vector basis of this subspace.
        """
        if not self.isBlade():
            raise ValueError("self is not a blade")

        # only one grade, since this is a blade
        gr, = self.grades()

        selfInv = self.inv()

        selfInv.clean()

        wholeBasis = []  # vector basis of the whole space

        for i in range(self.layout.gaDims):
            if self.layout.gradeList[i] == 1:
                v = np.zeros((self.layout.gaDims,), dtype=float)
                v[i] = 1.
                wholeBasis.append(self._newMV(v))

        thisBasis = []  # vector basis of this subspace

        J, mv = self._checkOther(1.)  # outer product of all of the vectors up
        # to the point of iteration

        for ei in wholeBasis:
            Pei = ei.lc(self) * selfInv

            J.clean()

            J2 = J ^ Pei

            if J2 != 0:
                J = J2
                thisBasis.append(Pei)
                if len(thisBasis) == gr:  # we have a complete set
                    break

        return thisBasis

    def join(self, other) -> 'MultiVector':
        r"""The join of two blades.

        :math:`J = A \wedge B`
        """

        other, mv = self._checkOther(other)

        grSelf = self.grades()
        grOther = other.grades()

        if len(grSelf) == len(grOther) == 1:
            # both blades
            grSelf, = grSelf
            grOther, = grOther

            # try the outer product first
            J = self ^ other

            if J != 0:
                return J.normal()

            # try something else
            M = (other * self.invPS()).lc(self)

            if M != 0:
                C = M.normal()
                J = (self * C.rightInv()) ^ other
                return J.normal()

            if grSelf >= grOther:
                A = self
                B = other
            else:
                A = other
                B = self

            if (A * B) == (A | B):
                # B is a subspace of A or the same if grades are equal
                return A.normal()

            # ugly, but general way
            # watch out for residues

            # A is still the larger-dimensional subspace

            Bbasis = B.basis()

            # add the basis vectors of B one by one to the larger
            # subspace except for the ones that make the outer
            # product vanish

            J = A

            for ei in Bbasis:
                J.clean()
                J2 = J ^ ei

                if J2 != 0:
                    J = J2

            # for consistency's sake, we'll normalize the join
            J = J.normal()

            return J

        else:
            raise ValueError("not blades")

    def meet(self, other, subspace=None) -> 'MultiVector':
        r"""The meet of two blades.

        Computation is done with respect to a subspace that defaults to
        the join if none is given.

        :math:`M \vee_i N = M_i^{-1}N`
        """

        other, mv = self._checkOther(other)

        r = self.grades()
        s = other.grades()

        if len(r) > 1 or len(s) > 1:
            raise ValueError("not blades")

        if subspace is None:
            subspace = self.join(other)

        return (self * subspace.inv()) | other

    def astype(self, *args, **kwargs):
        """
        Change the underlying scalar type of this vector.

        Can be used to force lower-precision floats or integers

        See `np.ndarray.astype` for argument descriptions.
        """
        return self._newMV(self.value.astype(*args, **kwargs))




import h5py
import numpy as np
import json


def write_json_file(file_name, mv_array, metric, basis_names, compression=True,
                    transpose=False, sparse=False, support=None, compression_opts=1):
    """
    Writes a json ga file of format version 0.0.1
    """
    data_dict = {}

    # Record the version number
    data_dict['version'] = '0.0.1'

    # First lets deal with the multivector coefficient data itself
    dset_data = {}
    if transpose:
        dset_data['data'] = mv_array.T.tolist()
        dset_data['transpose'] = True
    else:
        dset_data['data'] = mv_array.tolist()
        dset_data['transpose'] = False

    if sparse:
        dset_data['sparse'] = True
        if support is not None:
            dset_data['support'] = support.tolist()
        else:
            raise ValueError('You must specify the support of the multivectors '
                             'if you explicitly specify sparse storage')
    else:
        dset_data['sparse'] = False
        dset_data['support'] = []

    data_dict['dataset'] = dset_data

    # Now save the metric
    data_dict["metric"] = metric.tolist()

    # Now the basis names
    data_dict["basis_names"] = [str(s) for s in basis_names]

    with open(file_name, "w") as fp:
        json.dump(data_dict, fp)


def read_json_file(file_name):
    """
    Reads a json ga file of format version 0.0.1
    """
    with open(file_name, "r") as fp:
        f = json.load(fp)
        assert f['version'] == '0.0.1'
        data = f['dataset']
        transpose = data['transpose']
        if transpose:
            data_array = np.array(data['data']).T
        else:
            data_array = np.array(data['data'])
        sparse = data['sparse']
        if sparse:
            support = data['support']
        else:
            support = None
        metric = np.array(f['metric'])
        basis_names = np.array(f['basis_names'][:])
    return data_array, metric, basis_names, support


def write_ga_file(file_name, mv_array, metric, basis_names, compression=True,
                  transpose=False, sparse=False, support=None, compression_opts=1):
    """
    Writes a ga file of format version 0.0.1
    """
    with h5py.File(file_name, "w") as f:
        # Record the version number
        f.attrs['version'] = '0.0.1'

        # First lets deal with the multivector coefficient data itself
        if compression:
            if transpose:
                dset_data = f.create_dataset("data", data=mv_array.T, compression="gzip",
                                             compression_opts=compression_opts)
                dset_data.attrs['transpose'] = True
            else:
                dset_data = f.create_dataset("data", data=mv_array, compression="gzip",
                                             compression_opts=compression_opts)
                dset_data.attrs['transpose'] = False
        else:
            if transpose:
                dset_data = f.create_dataset("data", data=mv_array.T)
                dset_data.attrs['transpose'] = True
            else:
                dset_data = f.create_dataset("data", data=mv_array)
                dset_data.attrs['transpose'] = False

        if sparse:
            dset_data.attrs['sparse'] = True
            if support is not None:
                dset_support = f.create_dataset("support", support)
            else:
                raise ValueError('You must specify the support of the multivectors '
                                 'if you explicitly specify sparse storage')
        else:
            dset_data.attrs['sparse'] = False
            dset_support = f.create_dataset("support", data=np.array([], dtype=np.uint64))

        # Now save the metric
        dset_metric = f.create_dataset("metric", data=metric)

        # Now the basis names
        dset_basis_names = f.create_dataset("basis_names", data=basis_names)


def read_ga_file(file_name):
    """
    Reads a ga file of format version 0.0.1
    """
    with h5py.File(file_name, "r") as f:
        assert f.attrs['version'] == '0.0.1'
        data = f['data']
        transpose = data.attrs['transpose']
        if transpose:
            data_array = data[:].T
        else:
            data_array = data[:]
        sparse = data.attrs['sparse']
        if sparse:
            support = data.attrs['support']
        else:
            support = None
        metric = f['metric'][:]
        basis_names = f['basis_names'][:]
    return data_array, metric, basis_names, support



import unittest
from clifford import *
from clifford.g3c import *
from clifford.tools.g3c import *
from clifford.io import *


class TestParseMultivector(unittest.TestCase):

    def test_parse_multivector(self):
        A = layout.randomMV()
        B = layout.parse_multivector(str(A))
        np.testing.assert_almost_equal(A.value, B.value, 3)


class TestHDF5BasicIO(unittest.TestCase):

    def test_write_and_read(self):
        file_name = "test.ga"

        basis_names = np.array(list(sorted(layout.basis_vectors.keys())), dtype=bytes)

        mv_array = ConformalMVArray([random_point_pair() for i in range(1000)]).value
        write_ga_file(file_name, mv_array, layout.metric, basis_names, compression=True,
                      transpose=False, sparse=False, support=False)

        data_array, metric_2, basis_names_2, support = read_ga_file(file_name)

        np.testing.assert_equal(data_array, mv_array)
        np.testing.assert_equal(layout.metric, metric_2)
        np.testing.assert_equal(basis_names, basis_names_2)

    def test_write_and_read_array(self):
        file_name = "test.ga"

        mv_array = MVArray([random_point_pair() for i in range(1000)])
        mv_array.save(file_name, compression=True, transpose=False, sparse=False, support=False)

        loaded_array = layout.load_ga_file(file_name)

        np.testing.assert_equal(loaded_array.value, mv_array.value)


class TestJSONBasicIO(unittest.TestCase):

    def test_write_and_read(self):
        file_name = "test.ga.json"

        basis_names = np.array(list(sorted(layout.basis_vectors.keys())))

        mv_array = ConformalMVArray([random_point_pair() for i in range(1000)]).value
        write_json_file(file_name, mv_array, layout.metric, basis_names, compression=True,
                        transpose=False, sparse=False, support=False)

        data_array, metric_2, basis_names_2, support = read_json_file(file_name)

        np.testing.assert_equal(data_array, mv_array)
        np.testing.assert_equal(layout.metric, metric_2)
        np.testing.assert_equal(basis_names, basis_names_2)

    # def test_write_and_read_array(self):
    #     file_name = "test.ga.json"
    #
    #     mv_array = MVArray([random_point_pair() for i in range(1000)])
    #     mv_array.save(file_name, compression=True, transpose=False, sparse=False, support=False)
    #
    #     loaded_array = layout.load_ga_file(file_name)
    #
    #     np.testing.assert_equal(loaded_array.value, mv_array.value)


if __name__ == '__main__':
    unittest.main()



import random
from scipy import e
import numpy as np
import multiprocessing

from scipy.optimize import minimize

from .rotor_parameterisation import rotorconversion
from . import rotor_between_objects, apply_rotor, square_roots_of_rotor, rotor_between_lines, normalised
from clifford.g3c import *
from clifford.tools import orthoFrames2Versor as cartan
from clifford.tools.g3c import *
from .cost_functions import object_set_cost_sum, rotor_cost, val_rotor_cost_sparse
from clifford import grade_obj, MVArray

I3 = e123
I5 = e12345
imt_func = layout.imt_func
gmt_func = layout.gmt_func
inv_func = layout.inv_func
adjoint_func = layout.adjoint_func
e4_val = e4.value
ninf_val = einf.value
e123inf = e123*einf

motor_basis = [1 + 0 * e1, e12, e13, e23,
               e1 * einf, e2 * einf, e3 * einf,
               I3 * einf]

unit_rotor = np.zeros(32)
unit_rotor[0] = 1.0


@numba.njit
def val_two_line_rotor_estimate(LAval, LBval):
    """
    Get the rotor between two pairs of
    two oriented lines in space
    """
    Lp = val_normalised(neg_twiddle_root_val(LAval[0, :] + LAval[1, :])[0, :])
    Lm = val_normalised(neg_twiddle_root_val(LAval[0, :] - LAval[1, :])[0, :])
    Lp2 = val_normalised(neg_twiddle_root_val(LBval[0, :] + LBval[1, :])[0, :])
    Lm2 = val_normalised(neg_twiddle_root_val(LBval[0, :] - LBval[1, :])[0, :])
    R1 = val_rotor_between_objects_root(Lp, Lp2)
    R2 = val_rotor_between_objects_root(val_normalised(val_apply_rotor(Lm, R1)), Lm2)
    return val_normalised(gmt_func(R2, R1))


def two_line_rotor_estimate(LA, LB):
    """
    Get the rotor between two pairs of
    two oriented lines

    L1 = LA[0]
    L2 = LA[1]
    Lp = average_objects([L1, L2])
    Lm = average_objects([L1, -L2])
    L3 = LB[0]
    L4 = LB[1]
    Lp2 = average_objects([L3, L4])
    Lm2 = average_objects([L3, -L4])
    R1 = rotor_between_objects(Lp, Lp2)
    R2 = rotor_between_objects(normalised(R1*Lm*~R1), Lm2)
    return normalised(R2*R1)
    """
    return layout.MultiVector(value=val_two_line_rotor_estimate(
        np.array([l.value for l in LA]),
        np.array([l.value for l in LB])
    ))


def three_plane_rotor_estimate(PA, PB):
    """
        Get the rotor between two pairs of
        three oriented planes
    """
    return two_line_rotor_estimate(
        [normalised(meet(PA[0], PA[1])), normalised(meet(PA[1], PA[2]))],
        [normalised(meet(PB[0], PB[1])), normalised(meet(PB[1], PB[2]))])


def extract_rotor_from_TRS_mat_est(mat_est):
    """
    Given a matrix of the form [TRS_left@~TRS_right] returns TRS
    """
    sph = (up(e1)^up(-e1)^up(e2)^up(e3)).normal()*I5
    sph2 = layout.MultiVector(value=mat_est@sph.value).normal()
    t = down((sph2*einf*sph2)(1))
    T = generate_translation_rotor(t)
    S = generate_dilation_rotor(get_radius_from_sphere(sph2*I5)/get_radius_from_sphere(sph*I5))
    TS = T*S
    Rest = 1 + sum(
        (1 / layout.MultiVector(value=val_apply_rotor(mat_est@ebase.value, (~TS).value)))*ebase
        for ebase in [e1, e2, e3]
    )
    return (TS*Rest).normal()


def lambda_estimate(Y_point_list, X_point_list, n_samples=50):
    """
    Does a sample based lambda estimate as per Alex Arsenovic
    """
    lamb2est = 0
    for m in range(n_samples):
        indices = random.sample(range(len(X_point_list)), k=3)
        i = indices[0]
        j = indices[1]
        k = indices[2]
        lambdai2 = (X_point_list[i]|X_point_list[j])[0]
        lambdai2 *= (X_point_list[i]|X_point_list[k])[0]
        lambdai2 *= (Y_point_list[j]|Y_point_list[k])[0]
        denom = (Y_point_list[i]|Y_point_list[j])[0]
        denom *= (Y_point_list[i]|Y_point_list[k])[0]
        denom *= (X_point_list[j]|X_point_list[k])[0]
        lamb2est += lambdai2/denom
    return np.sqrt(lamb2est/n_samples)


def direct_TRS_extraction(Y, X):
    """
    Directly extracts the TRS rotor from vector correspondences
    """
    l = lambda_estimate(Y, X, n_samples=100)
    XArray = np.array([x.value for x in X]).T
    YArray = np.array([y.value for y in Y]).T * l
    lstsol = np.linalg.lstsq(XArray.T, YArray.T, rcond=None)[0].T
    est_rotor = extract_rotor_from_TRS_mat_est(lstsol)
    return est_rotor


@numba.njit
def val_project_bv(bv, x):
    return -imt_func(imt_func(x, bv), bv)


@numba.njit
def val_calc_error(R, A, Y):
    sumout = 0
    for i in range(A.shape[0]):
        sumout += val_norm(Y[i, :] - val_apply_rotor(A[i, :], R))
    return sumout/A.shape[0]


@numba.njit
def val_in_plane_estimate_rotation(bv, A, Y):
    rightside = np.zeros(32)
    leftside = np.zeros(32)
    for i in range(A.shape[0]):
        Abv = val_project_bv(bv, A[i, :])
        Ybv = val_project_bv(bv, Y[i, :])
        revAbv = adjoint_func(Abv)
        rightside += gmt_func(revAbv, Ybv)
        leftside += gmt_func(revAbv, Abv)
    R = gmt_func(inv_func(leftside), rightside)
    R[0] += 1
    return adjoint_func(val_normalised(R))


e23_val = e23.value
e12_val = e12.value
e13_val = e13.value
dekeninckbivmat = np.array([e23_val, e12_val, e13_val])


@numba.njit
def val_de_keninck_twist(Y, X, guess):
    """
    Performs the De Keninck twist
    Estimates the rotation between vectors
    """
    lasterr = 1E20
    nmvs = X.shape[0]
    A = np.zeros((nmvs, 32))
    for steps in range(1, 100):
        for i in range(nmvs):
            A[i, :] = val_apply_rotor(X[i, :], guess)
        biv = dekeninckbivmat[steps % 3, :]
        newguess = val_normalised(gmt_func(val_in_plane_estimate_rotation(biv, A, Y), guess))
        newerr = val_calc_error(newguess, X, Y)
        if newerr < lasterr:
            guess = newguess
        if abs(lasterr-newerr) < 1E-6:
            break
        lasterr = newerr
    return guess


def de_keninck_twist(Y, X, guess=None):
    """
    Performs the De Keninck twist
    Estimates the rotation between vectors
    """
    if guess is None:
        guess = (1.0 + 0*e1)
    return layout.MultiVector(value=val_de_keninck_twist(Y.value, X.value, guess.value))


def average_estimator(reference_model, query_model):
    """
    Estimates the rotor that takes one set of objects to another
    by estimating the rotor between the average objects
    """
    r_av = sum(reference_model)(grade_obj(reference_model[0]))
    q_av = sum(query_model)(grade_obj(reference_model[0]))
    return rotor_between_objects(q_av, r_av)


def estimate_rotor_objects_subsample(reference_model, query_model, n_repeats=None, objects_per_sample=None,
                                     maxfev=20000, print_res=False, pool_size=1,
                                     object_type='generic', motor=False):
    """
    Estimates the rotor that takes one set of objects to another
    """

    if n_repeats is None:
        n_repeats = int(len(query_model)/2)
    if objects_per_sample is None:
        objects_per_sample = max(int(len(query_model)/10), 5)
    if int(pool_size) < 2:
        min_cost = np.finfo(float).max
        min_rotor = 1.0 + 0.0 * e1
        for i in range(n_repeats):
            indices = random.sample(range(len(reference_model)), objects_per_sample)
            object_sample_reference = [reference_model[j] for j in indices]
            object_sample_query = [query_model[j] for j in indices]
            try:
                rotor, new_cost = estimate_rotor_objects(object_sample_reference, object_sample_query,
                                                         object_type=object_type, motor=motor)
            except Exception:
                rotor = 1.0 + 0*e1
                new_cost = np.inf
            if new_cost < min_cost:
                min_cost = new_cost
                min_rotor = rotor
            # print('SAMPLE: ', i, '  cost  ', min_cost)
    elif int(pool_size) > 1:
        def estimate_mot(Y, X):
            """ Alternate motor estimation func"""
            return estimate_rotor_objects(Y, X, motor=True)
        with multiprocessing.Pool(int(pool_size)) as pool_obj:
            object_sample_pairs = []
            for i in range(n_repeats):
                indices = random.sample(range(len(reference_model)), objects_per_sample)
                object_sample_reference = [reference_model[j] for j in indices]
                object_sample_query = [query_model[j] for j in indices]
                object_sample_pairs.append([object_sample_reference, object_sample_query])
            if object_type == 'lines':
                starmap_output = pool_obj.starmap(estimate_rotor_lines, object_sample_pairs)
            else:
                if motor:
                    starmap_output = pool_obj.starmap(estimate_mot, object_sample_pairs)
                else:
                    starmap_output = pool_obj.starmap(estimate_rotor_objects, object_sample_pairs)
            min_rotor, min_cost = min(starmap_output, key=lambda x: x[1])
            # print('SAMPLE: ', n_repeats, '  cost  ', min_cost)
    else:
        raise ValueError('Pool size is not valid')
    return min_rotor, min_cost


def dorst_motor_estimate(Q_in, P_in):
    """
    Estimating rotors from a variety of data
    """
    # Form linear operator
    Q_list = [Q(0, 1, 3) - Q(2, 4, 5) for Q in Q_in]
    P_list = P_in
    lag_mat = np.zeros((32, 32))
    n = len(Q_in)
    wk = 1 / n
    for k in range(n):
        Qk = Q_list[k]
        Qkrev = ~Qk
        Pk = P_list[k]
        Pkrev = ~Pk
        Qk_left = get_left_gmt_matrix(Qk.value)
        Pk_right = get_right_gmt_matrix(Pk.value)
        Qk_left_rev = get_left_gmt_matrix(Qkrev.value)
        Pk_right_rev = get_right_gmt_matrix(Pkrev.value)
        lag_mat += 0.5 * wk * (Qk_left @ Pk_right + Qk_left_rev @ Pk_right_rev)
    # Steps 2 and 3
    L = np.zeros((8, 8))
    for i in range(8):
        for j in range(8):
            L[i, j] = gmt_func((~motor_basis[i]).value, lag_mat @ motor_basis[j].value)[0]
    Lrr = L[0:4, 0:4]
    Lrq = L[0:4, 4:8]
    Lqr = L[4:8, 0:4]
    Lqq = L[4:8, 4:8]
    Lqqplus = np.linalg.pinv(Lqq)
    Ldash = Lrr - Lrq @ (Lqqplus @ Lqr)

    # Step 4
    w, v = np.linalg.eig(Ldash)
    max_e_val = 0.0
    n = 0
    for i in range(4):
        this_e_val = w[i]
        if this_e_val > max_e_val:
            max_e_val = this_e_val
            n = i
    r = v[:, n]

    # Step 5
    q = -((Lqqplus @ Lqr) @ r)
    # Step 6
    m = np.zeros(8)
    m[0:4] = r
    m[4:8] = q
    # Step 7
    M = sum((m[k] * motor_basis[k] for k in range(8))).normal()
    return M


def estimate_rotor_objects_subsample_sequential(reference_model, query_model, n_repeats=None, objects_per_sample=None,
                                                maxfev=20000, print_res=False, pool_size=1, object_type='generic'):
    """
    Estimates the rotor that takes one set of objects to another
    """
    if n_repeats is None:
        n_repeats = int(len(query_model)/2)
    if objects_per_sample is None:
        objects_per_sample = max(int(len(query_model)/10), 5)
    min_cost = np.finfo(float).max
    min_rotor = 1.0 + 0.0 * e1
    for i in range(n_repeats):
        indices = random.sample(range(len(reference_model)), objects_per_sample)
        object_sample_reference = [reference_model[j] for j in indices]
        object_sample_query = [query_model[j] for j in indices]
        try:
            rotor, e_flag = sequential_object_rotor_estimation(object_sample_reference, object_sample_query,
                                                               object_type=object_type)
        except Exception:
            rotor = 1.0 + 0*e1
            print('ERROR')
        query_model_remapped = [normalised(apply_rotor(l, rotor)) for l in object_sample_query]
        new_cost = object_set_cost_sum(reference_model, query_model_remapped, object_type=object_type)
        if new_cost < min_cost:
            min_cost = new_cost
            min_rotor = rotor
        # print('SAMPLE: ', i, '  cost  ', min_cost)
    return min_rotor, min_cost


def estimate_rotor_lines(reference_model, query_model, maxfev=20000, print_res=False):
    return estimate_rotor_objects(reference_model, query_model, maxfev=maxfev, print_res=print_res, object_type='lines')


def estimate_rotor_objects(reference_model, query_model, maxfev=20000,
                           print_res=False, object_type='generic', motor=False,
                           symmetric=False):
    """
    Estimates the rotor that takes one set of objects to another
    """
    grade_list = [grade_obj(q) for q in query_model]
    x0 = np.finfo(float).eps * np.random.rand(6)

    def minimisation_func(x):
        R = rotorconversion(x)
        query_model_remapped = [normalised((apply_rotor(l, R))(grade_list[i])) for i, l in enumerate(query_model)]
        return object_set_cost_sum(reference_model, query_model_remapped,
                                   object_type=object_type, motor=motor,
                                   symmetric=symmetric)

    res = minimize(minimisation_func, x0, method='SLSQP', options={'ftol': 10.0 ** (-16),
                                                                   'maxiter': 1000,
                                                                   'disp': False})
    if print_res:
        print(res)
    res = minimize(minimisation_func, res.x, method='L-BFGS-B', options={'ftol': 10.0**(-16),
                                                                         'maxiter': 1000,
                                                                         'disp': False,
                                                                         'maxls': 40})
    if print_res:
        print(res)
    rotor = rotorconversion(res.x)
    query_model_remapped = [normalised((apply_rotor(l, rotor))(grade_list[i])) for i, l in enumerate(query_model)]
    cost = object_set_cost_sum(reference_model, query_model_remapped, object_type=object_type, motor=motor,
                               symmetric=symmetric)
    return rotor, cost


def cartans_lines(obj_list_a, obj_list_b):
    """ Performs the extended cartans algorithm as suggested by Alex Arsenovic """
    V_found, rs = cartan(A=obj_list_a, B=obj_list_b)
    theta = ((V_found*~V_found)*e1234)(0)
    V_found = e**(-theta/2*e123inf)*V_found
    return V_found


def sequential_object_rotor_estimation(reference_model, query_model, n_iterations=500,
                                       cost_tolerance=10*(10**-16), random_sequence=False,
                                       object_type='generic', motor=True):
    """
    Performs a sequential rotor update based on the rotors between individual objects
    Exits when a full rotation through all objects produces a very small update of rotor
    """
    grade_list = [grade_obj(ob) for ob in query_model]

    R_total = 1.0 + 0.0*e1
    for j in range(n_iterations):
        r_set = 1.0 + 0.0*e1
        if random_sequence:
            indices = random.sample(range(len(query_model)), len(query_model))
        else:
            indices = range(len(query_model))
        for i in indices:
            grade = grade_list[i]
            new_obj = normalised(apply_rotor(query_model[i], R_total)(grade))
            C1 = normalised(new_obj)
            C2 = normalised(reference_model[i])
            if abs(C1 + C2) < 0.0001:
                C1 = -C1
            if object_type == 'lines':
                rroot = normalised(square_roots_of_rotor((rotor_between_objects(C1, C2)))[0])
            else:
                if motor:
                    if grade_obj(C1, 0.00001) == 4:
                        rroot = normalised(square_roots_of_rotor(rotor_between_objects(C1, C2))[0])
                    else:
                        rroot = normalised(square_roots_of_rotor(motor_between_objects(C1, C2))[0])
                else:
                    rroot = normalised(square_roots_of_rotor(rotor_between_objects(C1, C2))[0])
            r_set = normalised((rroot*r_set)(0, 2, 4))
            R_total = normalised((rroot * R_total)(0, 2, 4))
        if rotor_cost(r_set) < cost_tolerance:
            exit_flag = 0
            return R_total, exit_flag
    exit_flag = 1
    return R_total, exit_flag


@numba.njit
def set_as_unit_rotor_jit(array):
    for j in range(1, 32):
        array[j] = 0.0
    array[0] = 1.0


@numba.njit
def sequential_rotor_estimation_jit(reference_model, query_model, rotor_output,
                                    n_iterations=20, motor=True):
    cost_tolerance = 10 * (10 ** -16)

    # Allocate memory
    r_set = np.zeros(32)
    r_running = np.zeros(32)
    r_root = np.zeros(32)
    r_temp = np.zeros(32)
    C1 = np.zeros(32)

    # Set up the running rotor estimate
    set_as_unit_rotor_jit(r_running)

    # Start iterating for convergence
    for iteration_number in range(n_iterations):

        # Set up the convergence check
        set_as_unit_rotor_jit(r_set)

        # Iterate over the array of objects
        for mv_ind in range(reference_model.shape[0]):
            C1 = val_apply_rotor(query_model[mv_ind, :], r_running)
            C1 = val_normalised(C1)
            C2 = val_normalised(reference_model[mv_ind, :])

            if motor:
                C1_4 = project_val(C1, 4)
                if np.sum(np.abs(C1_4)) > 100*np.sum(np.abs(C1 - C1_4)):
                    r_temp = val_rotor_between_objects_root(C1, C2)
                else:
                    r_temp = val_motor_between_objects(C1, C2)
            else:
                r_temp = val_rotor_between_objects_root(C1, C2)
            r_temp[0] += 1.0
            r_root = pos_twiddle_root_val(r_temp)[0, :]

            # Update the set rotor and the running rotor
            r_temp = gmt_func(r_root, r_set)
            r_set = val_normalised(r_temp)
            r_temp = gmt_func(r_root, r_running)
            r_running = val_normalised(r_temp)

            # Check if we have converged
            if val_rotor_cost_sparse(r_set) < cost_tolerance:
                r_temp = val_normalised(r_running)
                for i in range(32):
                    rotor_output[i] = r_temp[i]

                # Now calculate the cost of this transform
                total_cost = 0.0
                for object_ind in range(query_model.shape[0]):
                    r_temp = val_apply_rotor(query_model[object_ind, :], rotor_output)
                    total_cost += np.abs(val_rotor_cost_sparse(val_rotor_between_objects_root(r_temp, reference_model[object_ind, :])))
                return total_cost
    # Return whatever we have
    r_temp = val_normalised(r_running)
    for i in range(32):
        rotor_output[i] = r_temp[i]

    total_cost = 0.0
    for object_ind in range(query_model.shape[0]):
        r_temp = val_apply_rotor(query_model[object_ind, :], rotor_output)
        total_cost += np.abs(val_rotor_cost_sparse(val_rotor_between_objects_root(r_temp, reference_model[object_ind, :])))
    return total_cost


@numba.njit
def sequential_rotor_estimation_chunks_jit(reference_model, query_model, output, cost_array):
    # Break the model into n chunks and estimate the rotor based on each of those
    n_chunks = output.shape[0]
    n_objects_per_chunk = int(reference_model.shape[0]/n_chunks)
    for i in range(n_chunks):
        ref = reference_model[i*n_objects_per_chunk:(i+1)*n_objects_per_chunk]
        qer = query_model[i*n_objects_per_chunk:(i+1)*n_objects_per_chunk]
        total_cost = sequential_rotor_estimation_jit(ref, qer, output[i, :])
        cost_array[i] = total_cost



import numpy as np


from clifford.io import write_ga_file, read_ga_file  # noqa: F401
from ._multivector import MultiVector

dual_array = np.vectorize(MultiVector.dual)
normal_array = np.vectorize(MultiVector.normal)
call_array = np.vectorize(MultiVector.__call__)


class MVArray(np.ndarray):
    '''
    MultiVector Array
    '''

    def __new__(cls, input_array):
        obj = np.empty(len(input_array), dtype=object)
        obj[:] = input_array
        obj = obj.view(cls)
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return

    @property
    def value(self):
        """
        Return an np array of the values of multivectors
        """
        return np.array([mv.value for mv in self])

    @staticmethod
    def from_value_array(layout, value_array):
        """
        Constructs an array of mvs from a value array
        """
        v_new_mv = np.vectorize(lambda v: MultiVector(layout, v), otypes=[MVArray], signature='(n)->()')
        return MVArray(v_new_mv(value_array))

    def save(self, filename, compression=True, transpose=False,
             sparse=False, support=False, compression_opts=1):
        """
        Saves the array to a ga file
        """
        write_ga_file(filename, self.value, self[0].layout.metric, self[0].layout.basis_names,
                      compression=compression, transpose=transpose,
                      sparse=sparse, support=support, compression_opts=compression_opts)

    def sum(self):
        '''
        sum elements of this MVArray
        '''
        out = self[0]
        for k in self[1:]:
            out += k
        return out

    def gp(self):
        '''
        geometric product of all elements of this MVArray  (like reduce)
        like `self[0]*self[1]*....self[n]`
        '''
        out = self[0]
        for k in self[1:]:
            out *= k
        return out

    def op(self):
        '''
        outer product of all elements of this MVArray  (like reduce)
        like `self[0]^self[1]^....self[n]`
        '''
        out = self[0]
        for k in self[1:]:
            out = out^k
        return out

    def normal(self):
        """
        Normalises all elements
        """
        return normal_array(self)

    def dual(self):
        """
        Takes the dual of all elements
        """
        return dual_array(self)

    def __call__(self, A):
        """
        Performs grade projection on all elements
        """
        return call_array(self, A)



'''
.. currentmodule:: clifford.cga

========================================
cga (:mod:`clifford.cga`)
========================================

Object Oriented Conformal Geometric Algebra.

Examples
-----------
>>> from clifford import Cl
>>> from clifford.cga import CGA
>>> g3, blades  = Cl(3)
>>> g3c = CGA(g3)
>>> C = g3c.round(3)             # create random sphere
>>> T = g3c.translation(e1+e2)   # create translation
>>> C_ = T(C)                    # translate the sphere
>>> C_.center                    # compute center of sphere

The CGA
========

.. autosummary::
    :toctree: generated/

    CGA


Objects
================

.. autosummary::
    :toctree: generated/

    Flat
    Round

Operators
================

.. autosummary::
    :toctree: generated/

    Rotation
    Dilation
    Translation
    Transversion

Meta-Class
===========

.. autosummary::
    :toctree: generated/

    CGAThing

'''

from functools import reduce
from . import conformalize, op, gp, MultiVector, Cl
from numpy import zeros, e, log
from numpy.random import rand
import math

try:
    import pyganja as ganja
    pyganja_available = True
except ImportError:
    pyganja_available = False


class CGAThing(object):
    '''
    base class for cga objects and operators.

    maps versor product to `__call__`.
    '''
    def __call__(self, other):
        if isinstance(other, MultiVector):
            if other.grades() == {1}:
                null = self.cga.null_vector(other)
                return self.mv*null*~self.mv
            return self.mv*other*~self.mv
        else:
            klass = other.__class__
            return klass(self.cga, self.mv*other.mv*~self.mv)

    def inverted(self):
        '''
        inverted version of this thing.

        self -> ep*self*ep

        where ep is the positive added basis vector
        '''
        return self.cga.ep * self.mv * self.cga.ep

    def involuted(self):
        '''
        inverted version of this thing.

        self -> E0*self*E0

        where E0 is the added minkowski bivector
        '''
        return self.cga.E0 * self.mv * self.cga.E0


# Objects
class Flat(CGAThing):
    '''
        A line, plane, or hyperplane.

        Typically constructed as method of existing cga, like `cga.flat()`

        multivector is accessable by `mv` property

        Parameters
        -----------
        cga : `CGA`
            the cga object

        args : [int, Multivector, Multivectors]
            * if nothing supplied, generate a flat of highest dimension
            * int: dimension of flat  (2=line, 3=plane, etc)
            * Multivector : can be
              * existing Multivector representing the Flat
              * vectors on the flat

        Examples
        ----------

        >>> cga.flat()               # from None
        >>> cga.flat(2)              # from dim of space
        >>> cga.flat(e1, e2)         # from points
        >>> cga.flat(cga.flat().mv)  # from existing multivector
        '''
    # could inherent some generic CGAObject class
    def __init__(self, cga, *args):

        self.cga = cga
        self.layout = cga.layout  # note: self.layout is the cga layout
        self.einf = self.cga.einf  # we use this alot

        if len(args) == 0:
            # generate random highest dimension flat
            nulls = [self.cga.null_vector() for k in range(self.layout.dims-2)]
            self.mv = reduce(op, nulls + [self.einf])

        elif len(args) == 1:
            # from existing multivector
            if isinstance(args[0], MultiVector):
                self.mv = args[0]

            # generate random flat  for given  dimension
            elif isinstance(args[0], int):
                dim = args[0]
                points = [self.cga.base_vector() for k in range(dim+1)]
                points = list(map(self.cga.up, points))
                self.mv = reduce(op, points + [self.einf])

        # from vectors on flat
        else:
            nulls = map(self.cga.null_vector, args)
            if self.einf not in nulls:
                nulls = list(nulls)+[self.einf]

            self.mv = reduce(op, nulls)

        self.mv = self.mv.normal()

    def __repr__(self):
        return '%i-Flat' % self.dim


class Round(CGAThing):
    '''
    A point pair, circle, sphere or hyper-sphere.

    Typically constructed as method of existing cga, like `cga.round()`

    multivector is accessable by `mv` property

    Parameters
    -----------
    cga : `CGA`
        the cga object

    args : [int, Multivector, Multivectors]
        * if nothing supplied, generate a round of highest dimension
        * int: dimension of flat  (2=point pair, 3=circle, etc)
        * Multivector : can be
          * existing Multivector representing the round
          * vectors on the round

    Examples
    ----------

    >>> cga.round()               # from None
    >>> cga.round(2)              # from dim of space
    >>> cga.round(e1, e2, -e1)    # from points
    >>> cga.round(cga.flat().mv)  # from existing multivector
    '''
    # could inherent some generic CGAObject class
    def __init__(self, cga, *args):
        # me
        self.cga = cga
        self.layout = cga.layout  # note: self.layout is the cga layout

        if len(args) == 0:
            # generate random highest dimension round
            nulls = [self.cga.null_vector() for k in range(self.layout.dims-1)]
            self.mv = reduce(op, nulls)

        elif len(args) == 1:
            # from existing multivector
            if isinstance(args[0], MultiVector):
                self.mv = args[0]

            # generate random round  for given  dimension
            elif isinstance(args[0], int):
                dim = args[0]
                points = [self.cga.base_vector() for k in range(dim+2)]
                points = map(self.cga.up, points)
                self.mv = reduce(op, points)

            # from center, radius tuple
            else:
                if len(args[0]) == 2:
                    center, radius = args[0]
                    center = self.cga.null_vector(center)
                    dual_round = (center - .5*radius**2*self.cga.einf)
                    self.mv = dual_round.normal().dual()

        # from vectors on round
        else:
            nulls = map(self.cga.null_vector, args)
            self.mv = reduce(op, nulls)

        self.mv = self.mv.normal()

    def from_center_radius(self, center, radius):
        '''
        construct a round from center/radius
        '''
        center = self.cga.null_vector(center)
        self.mv = (center - .5*radius**2*self.cga.einf).normal().dual()
        return self

    def __repr__(self):
        names = {4: 'Sphere', 3: 'Circle', 2: 'Point Pair', 1: 'Point'}
        if self.dim <= 4:
            return names[self.dim + 2]
        else:
            return '%i-Round' % (self.dim + 2)

    @property
    def dim(self):
        '''
        dimension of this round
        '''
        gr, = self.mv.grades()
        return gr - 2

    @property
    def center(self):
        '''
        center of this round, as a null vector
        '''
        return self.mv * self.cga.einf * self.mv

    @property
    def center_down(self):
        '''
        center of this round, as a down-projected  vector (in I_base)

        (but still in cga's layout)
        '''
        return self.cga.down(self.center)

    @property
    def radius(self):
        '''
        radius of the round (a float)
        '''
        dual_sphere = self.dual
        dual_sphere / (-dual_sphere | self.cga.einf)
        return math.sqrt(abs(dual_sphere * dual_sphere))

    @property
    def dual(self):
        '''
        self.mv* self.layout.I
        '''
        return self.mv * self.layout.I


# Operators
class Translation(CGAThing):
    '''
    A Translation

    Can be constructed from a vector in base space or a null
    vector, or nothing.

    Parameters
    ----------
    args : [none, `clifford.Multivector`]
        if none, a random translation will be generated
        several types of Multivectors can be used:

         * base vector - vector in base space
         * null vector
         * existing translation rotor

    Examples
    ----------
    >>> T = cga.translation()       # from None
    >>> T = cga.translation(e1+e2)  # from base vector
    >>> T = cga.translation(cga.up(e1+e2)) # from null vector
    >>> T = cga.translation(T.mv)  # from existing translation rotor
    '''
    def __init__(self, cga, *args):
        self.cga = cga
        self.layout = cga.layout

        if len(args) == 0:
            # generate generator!
            mv = 1 - self.cga.base_vector()*self.cga.einf/2.

        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, MultiVector):
                if arg.grades() == {1}:
                    # we have vector
                    mv = 1 - self.cga.straight_up(arg)*self.cga.einf/2.
                if arg.grades() == {0, 2}:
                    # we have ro tor
                    # TODO ensure its a translation
                    mv = args[0]
        else:
            raise ValueError('bad input')

        self.mv = mv

    def __repr__(self):
        return 'Translation'


class Dilation(CGAThing):
    '''
    A global dilation

    Parameters
    ----------
    args : [none, number]
        if none, a random dilation will be generated
        if a number, dilation of given amount

    Examples
    ----------
    >>> D = cga.dilation()          # from  none
    >>> D = cga.dilation(.4)        # from number
    '''
    def __init__(self, cga, *args):
        self.cga = cga
        self.layout = cga.layout

        if len(args) == 0:
            # generate a dilation
            args = [rand()]

        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, MultiVector):
                if arg.grades() == {0, 2}:
                    # we have a rotor
                    mv = arg
                if arg.grades == {0}:
                    arg = float(arg)

            if arg < 0:
                raise(ValueError('dilation should be positive'))

            mv = e**((-log(arg)/2.)*(self.cga.E0))

        else:
            raise ValueError('bad input')

        self.mv = mv

    def __repr__(self):
        return 'Dilation'


class Rotation(CGAThing):
    '''
    A Rotation

    Can be constructed from a generator, rotor, or none

    Parameters
    ----------
    args : [none, `clifford.Multivector`]
        if none, a random translation will be generated
        several types of Multivectors can be used:

         * bivector   - interpreted as the generator
         * existing translation rotor

    Examples
    ----------
    >>> R = cga.rotation()          # from None
    >>> R = cga.rotation(e12+e23)   # from bivector
    >>> R = cga.rotation(R.mv)   # from bivector
    '''
    def __init__(self, cga, *args):
        self.cga = cga
        self.layout = cga.layout

        if len(args) == 0:
            # generate a rotation

            U = self.layout.randomMV()(2)
            U = self.cga.I_base.project(U)
            self.mv = e**(U)

        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, MultiVector):
                if arg.grades() == {0, 2}:
                    # we have a rotor
                    self.mv = arg

                elif arg.grades() == {2}:
                    # we have a bivector, make sure its in base space
                    if arg^self.cga.I_base != 0:
                        arg = self.cga.I_base.project(arg)
                    self.mv = e**(arg)
                else:
                    # multivector has improper grade
                    raise ValueError('bad input')
            else:
                # arg isnt a multivector
                raise ValueError('bad input')

        else:
            # more than 1 arg
            raise ValueError('bad input')

    def __repr__(self):
        return 'Rotation'


class Transversion(Translation):
    '''
    A Transversion

    A transversion is a combination of an inversion-translation-inversion,
    or in other words an inverted translation operator. This inherits
    from `Translation`

    Can be constructed from a vector in base space or a null
    vector, or nothing.

    Parameters
    ----------
    args : [none, `clifford.Multivector`]
        if none, a random transversion will be generated
        several types of Multivectors can be used:

         * base vector - vector in base space
         * null vector
         * existing transversion rotor

    Examples
    ----------
    >>> K = cga.transversion()       # from None
    >>> K = cga.transversion(e1+e2)  # from base vector
    >>> K = cga.transversion(cga.up(e1+e2)) # from null vector
    >>> K = cga.transversion(T.mv)  # from existing translation rotor
    '''
    def __init__(self, cga, *args):
        self.mv = Translation(cga, *args).inverted()

    def __repr__(self):
        return 'Transversion'


class CGA(object):
    '''
    Conformal Geometric Algebra

    conformalizes the layout_orig,  and provides several
    methods and for objects/operators

    Parameters
    -----------
    layout_orig: [`clifford.Layout`, int]
        a layout for the *base* geometric algebra which is conformalized
        if given as an int, then generates a euclidean space of given
        dimension

    Examples
    ----------
    >>> from clifford import Cl
    >>> from clifford.cga import CGA
    >>> g3, blades  = Cl(3)
    >>> g3c = CGA(g3)
    >>> g3c = CGA(3)

    '''
    def __init__(self, layout_orig):
        if isinstance(layout_orig, int):
            layout_orig, blades = Cl(layout_orig)
        self.layout_orig = layout_orig
        self.layout, self.blades, stuff = conformalize(layout_orig)
        self.__dict__.update(stuff)

    # Objects
    def base_vector(self):
        '''
        random vector in the lower(original) space
        '''
        return self.I_base.project(self.layout.randomV())

    def null_vector(self, x=None):
        '''
        generates random null vector if x is None, or
        returns a null vector from base vector x, if  x^self.I_base ==0
        returns x,

        a null vector will lay on the horisphere
        '''
        if x is None:
            return self.up(self.base_vector())
        else:
            if x^self.I_base == 0:
                return self.up(x)
            return x

    def round(self, *args):
        '''
        see :class:`Round`
        '''
        return Round(self, *args)

    def flat(self, *args):
        '''
        see :class:`Flat`
        '''
        return Flat(self, *args)

    #  Operators
    def translation(self, *args):
        '''
        see :class:`Translation`
        '''
        return Translation(self, *args)

    def transversion(self, *args):
        '''
        see :class:`Transversion`
        '''
        return Transversion(self, *args)

    def dilation(self, *args):
        '''
        see :class:`Dilation`
        '''
        return Dilation(self, *args)

    def rotation(self, *args):
        '''
        see :class:`Rotation`
        '''
        return Rotation(self, *args)

    #  methods
    def straight_up(self, x):
        '''
        place a vector from layout_orig into this CGA, without up()
        '''
        return self.I_base.project(self.up(x))




from __future__ import print_function
from clifford import grade_obj
from clifford.g3c import *
from clifford.tools.g3c import interpret_multivector_as_object
from clifford.io import read_ga_file
from clifford import MVArray
import json


def interpolate_colors(alpha, rgb_a, rgb_b):
    rgb_c = (1-alpha)*rgb_a + alpha*rgb_b
    return 'rgb(' + str(int(rgb_c[0])) + ',' + str(int(rgb_c[1])) + ',' + str(int(rgb_c[2])) + ')'


class Sphere():
    def __init__(self, mv, color):
        self.mv = mv
        self.color = color

    def gen_string(self):
        return_string = 'DrawSphere('+str(self.mv)+','+self.color+');'
        return return_string


class Plane():
    def __init__(self, mv, color):
        self.mv = mv
        self.color = color

    def gen_string(self):
        return_string = 'DrawPlane('+str(self.mv)+','+self.color+');'
        return return_string


class Line():
    def __init__(self, mv, color):
        self.mv = mv
        self.color = color

    def gen_string(self):
        return_string = 'DrawLine('+str(self.mv)+','+self.color+');'
        return return_string


class Circle():
    def __init__(self, mv, color):
        self.mv = mv
        self.color = color

    def gen_string(self):
        return_string = 'DrawCircle('+str(self.mv)+','+self.color+');'
        return return_string


class PointPair():
    def __init__(self, mv, color):
        self.mv = mv
        self.color = color

    def gen_string(self):
        return_string = 'DrawPointPair('+str(self.mv)+','+self.color+');'
        return return_string


class EucPoint():
    def __init__(self, mv, color):
        self.mv = mv
        self.color = color

    def gen_string(self):
        return_string = 'DrawEucPoint('+str(self.mv)+','+self.color+');'
        return return_string


class ConfPoint():
    def __init__(self, mv, color):
        self.mv = mv
        self.color = color

    def gen_string(self):
        return_string = 'DrawConfPoint('+str(self.mv)+','+self.color+');'
        return return_string


class GAScene():

    def __init__(self):
        self.objects = []

    def add(self, object_added):
        self.objects.append(object_added)

    def concat(self, other):
        if isinstance(other, GAScene):
            for obj in other.objects:
                self.add(obj)
        else:
            raise ValueError('Other object must be a GAScene')

    def clear(self):
        self.objects = []

    def __repr__(self):
        s = ''
        for obj in self.objects:
            s += obj.gen_string() + '\n'
        s = s[:-1]
        return s

    def add_euc_point(self, mv, color='rgb(0,0,0)'):
        if grade_obj(mv) != 1:
            raise ValueError('Input is not a euc_point')
        self.add(EucPoint(mv, color))

    def add_conf_point(self, mv, color='rgb(0,0,0)'):
        if grade_obj(mv) != 1:
            raise ValueError('Input is not a conf_point')
        self.add(ConfPoint(mv, color))

    def add_circle(self, mv, color='rgb(0,0,0)'):
        if grade_obj(mv) != 3:
            raise ValueError('Input is not a circle')
        self.add(Circle(mv, color))

    def add_line(self, mv, color='rgb(0,0,0)'):
        if grade_obj(mv) != 3:
            raise ValueError('Input is not a line')
        self.add(Line(mv, color))

    def add_plane(self, mv, color='rgb(0,0,0)'):
        if grade_obj(mv) != 4:
            raise ValueError('Input is not a plane')
        self.add(Plane(mv, color))

    def add_sphere(self, mv, color='rgb(0,0,0)'):
        if grade_obj(mv) != 4:
            raise ValueError('Input is not a sphere')
        self.add(Sphere(mv, color))

    def add_point_pair(self, mv, color='rgb(0,0,0)'):
        if grade_obj(mv) != 2:
            raise ValueError('Input is not a point_pair')
        self.add(PointPair(mv, color))

    def add_object(self, mv, mv_type='interp', color='rgb(0,0,0)'):
        if mv_type == 'line':
            self.add_line(mv, color)
        elif mv_type == 'conf_point':
            self.add_conf_point(mv, color)
        elif mv_type == 'euc_point':
            self.add_euc_point(mv, color)
        elif mv_type == 'point_pair':
            self.add_point_pair(mv, color)
        elif mv_type == 'circle':
            self.add_circle(mv, color)
        elif mv_type == 'plane':
            self.add_plane(mv, color)
        elif mv_type == 'sphere':
            self.add_sphere(mv, color)
        elif mv_type == 'interp':
            """
            -1 -> not a blade
            0 -> a 1 vector but not a point
            1 -> a euclidean point
            2 -> a conformal point
            3 -> a point pair
            4 -> a circle
            5 -> a line
            6 -> a sphere
            7 -> a plane
            """
            return_val = interpret_multivector_as_object(mv)
            if return_val > 0:
                if return_val == 1:
                    self.add_euc_point(mv, color=color)
                elif return_val == 2:
                    self.add_conf_point(mv, color=color)
                elif return_val == 3:
                    self.add_point_pair(mv, color=color)
                elif return_val == 4:
                    self.add_circle(mv, color=color)
                elif return_val == 5:
                    self.add_line(mv, color=color)
                elif return_val == 6:
                    self.add_sphere(mv, color=color)
                elif return_val == 7:
                    self.add_plane(mv, color=color)
            else:
                raise ValueError('Input object cannot be interpretted as an object.')
        else:
            raise ValueError(str(mv_type) + ' is not a valid mv_type. You must specify a valid mv_type.')

    def add_object_array(self, mv_array, mv_type='interp', color='rgb(0,0,0)'):
        for mv in mv_array:
            self.add_object(mv, mv_type, color=color)

    def save_to_file(self, filename):
        with open(filename, 'w') as fobj:
            print(self, file=fobj)


class GanjaScene:

    def __init__(self):
        self.internal_list = []

    def add_object(self, mv, color=int('AA000000', 16), label=None):
        if label is not None:
            try:
                assert isinstance(label, str)
                self.internal_list.append(label)
            except:
                raise ValueError('Labels must be strings')
        self.internal_list.append(color)
        self.internal_list.append([i for i in mv.value])

    def add_objects(self, mv_list, color=int('AA000000', 16), label=None):
        if label is not None:
            try:
                assert isinstance(label, str)
                self.internal_list.append(label)
            except:
                raise ValueError('Labels must be strings')
        self.internal_list.append(color)
        for mv in mv_list:
            self.internal_list.append([i for i in mv.value])

    def __repr__(self):
        return json.dumps(self.internal_list)

    def save_to_file(self, filename):
        with open(filename, 'w') as fobj:
            print(self, file=fobj)


def draw_objects_ganja(objects, color=int('AA000000', 16), print_scene=True):
    """
    Takes a list of multivectors or a .ga file name and draws the multivectors
    By default attempts to interpret the type of object unless a mv_type is specified
    """
    if isinstance(objects, str):
        data_array, metric, basis_names, support = read_ga_file(objects)
        mv_list = [layout.MultiVector(value=data_array[i, :]) for i in range(data_array.shape[0])]
        sc = GanjaScene()
        sc.add_objects(mv_list, color=color)
        if print_scene:
            print(sc)
        return sc
    elif isinstance(objects, list) or isinstance(objects, MVArray):
        sc = GanjaScene()
        sc.add_objects(objects, color=color)
        if print_scene:
            print(sc)
        return sc
    else:
        raise ValueError('The input is not a string or a list of objects')


def draw_objects(objects, mv_type='interp', color='rgb(0,0,0)', print_scene=True):
    """
    Takes a list of multivectors or a .ga file name and draws the multivectors
    By default attempts to interpret the type of object unless a mv_type is specified
    """
    if isinstance(objects, str):
        data_array, metric, basis_names, support = read_ga_file(objects)
        mv_list = [layout.MultiVector(value=data_array[i, :]) for i in range(data_array.shape[0])]
        sc = GAScene()
        sc.add_object_array(mv_list, mv_type, color=color)
        if print_scene:
            print(sc)
        return sc
    elif isinstance(objects, list) or isinstance(objects, MVArray):
        sc = GAScene()
        sc.add_object_array(objects, mv_type, color=color)
        if print_scene:
            print(sc)
        return sc
    else:
        raise ValueError('The input is not a string or a list of objects')


class GAAnimation():
    def __init__(self):
        self.scenes = []

    def add_scene(self, scene):
        self.scenes.append(scene)

    def __repr__(self):
        s = ''
        for obj in self.scenes:
            s += str(obj).replace("\n", "#") + '\n'
        s = s[:-1]
        return s

    def save_to_file(self, filename):
        with open(filename, 'w') as fobj:
            print(self, file=fobj)



"""
Tools for 3DCGA (g3c)

3DCGA Tools
==========================================================

Generation Methods
--------------------

.. autosummary::
    :toctree: generated/

    random_bivector
    standard_point_pair_at_origin
    random_point_pair_at_origin
    random_point_pair
    standard_line_at_origin
    random_line_at_origin
    random_line
    random_circle_at_origin
    random_circle
    random_sphere_at_origin
    random_sphere
    random_plane_at_origin
    random_plane

    generate_n_clusters
    generate_random_object_cluster
    random_translation_rotor
    random_rotation_translation_rotor
    random_conformal_point
    generate_dilation_rotor
    generate_translation_rotor


Geometry Methods
--------------------

.. autosummary::
    :toctree: generated/

    intersect_line_and_plane_to_point
    val_intersect_line_and_plane_to_point
    quaternion_and_vector_to_rotor
    get_center_from_sphere
    get_radius_from_sphere
    point_pair_to_end_points
    val_point_pair_to_end_points
    get_circle_in_euc
    circle_to_sphere
    line_to_point_and_direction
    get_plane_origin_distance
    get_plane_normal
    get_nearest_plane_point
    val_convert_2D_polar_line_to_conformal_line
    convert_2D_polar_line_to_conformal_line
    val_convert_2D_point_to_conformal
    convert_2D_point_to_conformal
    val_distance_point_to_line
    distance_polar_line_to_euc_point_2d
    midpoint_between_lines
    val_midpoint_between_lines
    midpoint_of_line_cluster
    val_midpoint_of_line_cluster
    val_midpoint_of_line_cluster_grad
    get_line_intersection
    val_get_line_intersection
    project_points_to_plane
    project_points_to_sphere
    project_points_to_circle
    project_points_to_line

Misc
--------------------

.. autosummary::
    :toctree: generated/


    meet_val
    meet
    normalise_n_minus_1
    val_normalise_n_minus_1
    val_apply_rotor
    apply_rotor
    val_apply_rotor_inv
    apply_rotor_inv
    euc_dist
    mult_with_ninf
    val_norm
    norm
    val_normalised
    normalised
    val_up
    fast_up
    val_normalInv
    val_homo
    val_down
    fast_down
    dual_func
    fast_dual
    disturb_object
    project_val
    get_line_reflection_matrix
    val_get_line_reflection_matrix
    val_truncated_get_line_reflection_matrix
    interpret_multivector_as_object
    normalise_TR_to_unit_T
    scale_TR_translation


Root Finding
--------------------

.. autosummary::
    :toctree: generated/

    dorst_norm_val
    check_sigma_for_positive_root_val
    check_sigma_for_positive_root
    check_sigma_for_negative_root_val
    check_sigma_for_negative_root
    check_infinite_roots_val
    check_infinite_roots
    positive_root_val
    negative_root_val
    positive_root
    negative_root
    general_root_val
    general_root
    val_annihilate_k
    annihilate_k
    pos_twiddle_root_val
    neg_twiddle_root_val
    pos_twiddle_root
    neg_twiddle_root
    square_roots_of_rotor
    n_th_rotor_root
    interp_objects_root
    general_object_interpolation
    average_objects
    val_average_objects_with_weights
    val_average_objects
    rotor_between_objects
    val_rotor_between_objects_root
    val_rotor_between_objects_explicit
    calculate_S_over_mu
    val_rotor_between_lines
    rotor_between_lines
    rotor_between_planes
    val_rotor_rotor_between_planes

"""

import math
import numba
import numpy as np
from clifford.tools.g3 import quaternion_to_rotor, random_euc_mv, \
    random_rotation_rotor, generate_rotation_rotor, val_random_euc_mv
from clifford.g3c import *
import clifford as cf
from clifford import grades_present, NUMBA_PARALLEL, MVArray
from scipy.interpolate import interp1d

try:
    from pyganja import draw
    pyganja_available = True
except ImportError:
    pyganja_available = False


# Allow syntactic alternatives to the standard included in the clifford package
ninf = einf
no = -eo

# Define some useful objects
E = ninf ^ (no)
I5 = e12345
I3 = e123
E0 = ninf ^ -no
niono = ninf ^ no
E0_val = E0.value
I5_val = I5.value
ninf_val = ninf.value
no_val = no.value
I3_val = I3.value
eo_val = eo.value

unit_scalar_mv = 1.0 + 0.0*e1
unit_scalar_mv_val = unit_scalar_mv.value

adjoint_func = layout.adjoint_func
gmt_func = layout.gmt_func
omt_func = layout.omt_func
imt_func = layout.imt_func

epsilon = 10*10**(-6)


mask0 = layout.get_grade_projection_matrix(0)
mask1 = layout.get_grade_projection_matrix(1)
mask2 = layout.get_grade_projection_matrix(2)
mask3 = layout.get_grade_projection_matrix(3)
mask4 = layout.get_grade_projection_matrix(4)
mask5 = layout.get_grade_projection_matrix(5)
mask_2minus4 = mask2 - mask4


def interpret_multivector_as_object(mv):
    """
    Takes an input multivector and returns what kind of object it is
    -1 -> not a blade
    0 -> a 1 vector but not a point
    1 -> a euclidean point
    2 -> a conformal point
    3 -> a point pair
    4 -> a circle
    5 -> a line
    6 -> a sphere
    7 -> a plane
    """
    g_pres = grades_present(mv, 0.00000001)
    if len(g_pres) != 1:  # Definitely not a blade
        return -1
    grade, = g_pres
    if grade == 1:
        # This can now either be a euc mv or a conformal point or something else
        if mv(e123) == mv:
            return 1  # Euclidean point
        elif np.sum(np.abs((mv**2).value)) < 0.00000001:
            return 2  # Conformal point
        else:
            return 0  # Unknown mv
    elif mv.isBlade():
        if grade == 2:
            return 3  # Point pair
        elif grade == 3:  # Line or circle
            if abs(mv[e123]) > epsilon:
                return 4  # Circle
            else:
                return 5  # Line
        elif grade == 4:  # Sphere or plane
            if abs(((mv*I5)|no)[0]) > epsilon:
                return 7  # Plane
            else:
                return 6  # Sphere
    else:
        return -1


@numba.njit
def val_sphere_line_intersect(s, l):
    """
    Checks for intersection between a sphere and a line
    """
    mv = meet_val(s, l)
    return imt_func(mv, mv)[0] > 0


def sphere_line_intersect(s, l):
    """
    Checks for intersection between a sphere and a line
    """
    return val_sphere_line_intersect(s.value, l.value)


def sphere_in_sphere(S1, S2, tolerance=10**-6):
    """
    Checks if one sphere is inside the other
    (S1|S2)[0] < -1
    """
    return (unsign_sphere(S1)|unsign_sphere(S2))[0] <= -1 + tolerance


def sphere_beyond_plane(sphere, plane):
    """
    Check if the sphere is fully beyond the plane in the direction of
    the plane normal
    """
    snorm = unsign_sphere(sphere)
    return (snorm|plane)[0] < -get_radius_from_sphere(snorm)


def sphere_behind_plane(sphere, plane):
    """
    Check if the sphere is fully behind the plane in the direction of
    the plane normal, ie the opposite of sphere_beyond_plane
    """
    snorm = unsign_sphere(sphere)
    return (snorm|plane)[0] > get_radius_from_sphere(snorm)


def point_beyond_plane(point, plane):
    """
    Check if the point is fully beyond the plane in the direction of
    the plane normal
    """
    return (point|(I5*plane))[0] < 0


def unsign_sphere(S):
    """
    Normalises the sign of a sphere
    """
    return -normalised(S * ((I5 * S) | einf)[0])


def val_unsign_sphere(S):
    """
    Normalises the sign of a sphere
    """
    return val_normalised(S*(-imt_func(dual_func(S), ninf_val)[0]))


def join_spheres(S1in, S2in):
    """
    Find the smallest sphere that encloses both spheres
    """
    s1 = unsign_sphere(S1in)
    s2 = unsign_sphere(S2in)
    L = normalised(((s1 * I5) ^ (s2 * I5) ^ einf)(3))
    pp1 = normalised(meet(s1, L)(2))
    pp2 = normalised(meet(s2, L)(2))
    p1 = point_pair_to_end_points(pp1)[0]
    p2 = point_pair_to_end_points(pp2)[1]
    if (p1|(s2*I5))[0] > 0.0:
        opt_sphere = s2(4)
    elif (p2|(s1*I5))[0] > 0.0:
        opt_sphere = s1(4)
    else:
        p12 = p1 ^ p2
        L2 = normalised(p12 * (p12 ^ einf))
        opt_sphere = (L2*I5)(4)
    return unsign_sphere(opt_sphere)


def enclosing_sphere(spheres):
    """
    For a given set of spheres this finds a sphere that encloses all of them
    Tries to find the smallest one it can
    """
    nspheres = len(spheres)
    if nspheres == 1:
        return spheres[0]
    elif nspheres == 2:
        return join_spheres(spheres[0], spheres[1])
    mins = spheres[0]
    for i in range(1, nspheres):
        mins = join_spheres(mins, spheres[i])
    return mins


def project_points_to_plane(point_list, plane):
    """
    Takes a load of points and projects them onto a plane
    """
    projected_list = []
    for point in point_list:
        proj_point = ((point|plane)*plane)
        proj_point = normalise_n_minus_1((proj_point*einf*proj_point)(1))
        projected_list.append(proj_point)
    return projected_list


def project_points_to_sphere(point_list, sphere, closest=True):
    """
    Takes a load of points and projects them onto a sphere
    """
    if closest:
        point_index = 1
    else:
        point_index = 0
    projected_list = []
    C = sphere*einf*sphere
    for point in point_list:
        proj_point = point_pair_to_end_points(meet(normalised(point^C^einf), sphere))[point_index]
        projected_list.append(proj_point)
    return projected_list


def project_points_to_circle(point_list, circle, closest=True):
    """
    Takes a load of point and projects them onto a circle
    The closest flag determines if it should be the closest or furthest point on the circle
    """
    circle_plane = normalised(circle^einf)
    planar_points = project_points_to_plane(point_list, circle_plane)
    circle_points = project_points_to_sphere(planar_points, -circle*circle_plane*I5, closest=closest)
    return circle_points


def project_points_to_line(point_list, line):
    """
    Takes a load of points and projects them onto a line
    """
    projected_list = []
    for point in point_list:
        pp = point|line
        proj_point = (pp*einf*pp)(1)
        projected_list.append(proj_point)
    return projected_list


def closest_points_on_circles(C1, C2, niterations=20):
    """
    Given two circles C1 and C2 this calculates the closest
    points on each of them to the other
    """
    cav = average_objects([C1, C2])
    cav2 = average_objects([C1, -C2])
    PP = meet(cav, cav2^einf).normal()
    P_list = point_pair_to_end_points(PP)
    dmin = np.inf
    for Ptest in P_list:
        d = -(project_points_to_circle([Ptest], C1)[0](1)|project_points_to_circle([Ptest], C2)[0](1))[0]
        if d < dmin:
            dmin = d
            P2 = Ptest
    P1 = project_points_to_circle([P2], C1)[0](1)
    P2 = project_points_to_circle([P1], C2)[0](1)
    for i in range(niterations):
        P1 = project_points_to_circle([P2], C1)[0](1)
        P2 = project_points_to_circle([P1], C2)[0](1)
    return P1, P2


def closest_points_circle_line(C, L, niterations=20):
    """
    Given a circle C and line L this calculates the closest
    points on each of them to the other
    """
    cav = average_objects([C, L])
    cav2 = average_objects([C, -L])
    PP = meet(cav, cav2^einf).normal()
    P_list = point_pair_to_end_points(PP)
    dmin = np.inf
    for Ptest in P_list:
        d = -(project_points_to_circle([Ptest], C)[0](1)|project_points_to_line([Ptest], L)[0](1))[0]
        if d < dmin:
            dmin = d
            P2 = Ptest
    P1 = project_points_to_circle([P2], C)[0](1)
    P2 = project_points_to_line([P1], L)[0](1)
    for i in range(niterations):
        P1 = project_points_to_circle([P2], C)[0](1)
        P2 = project_points_to_line([P1], L)[0](1)
    return P1, P2


def furthest_points_on_circles(C1, C2, niterations=20):
    """
    Given two circles C1 and C2 this calculates the closest
    points on each of them to the other
    """
    P2 = random_conformal_point()
    P1 = project_points_to_circle([P2], C1, closest=False)[0](1)
    P2 = project_points_to_circle([P1], C2, closest=False)[0](1)
    for i in range(niterations):
        P1 = project_points_to_circle([P2], C1, closest=False)[0](1)
        P2 = project_points_to_circle([P1], C2, closest=False)[0](1)
    return P1, P2


def normalise_TR_to_unit_T(TR):
    """
    Takes in a TR rotor
    extracts the R and T
    normalises the T to unit displacement magnitude
    rebuilds the TR rotor with the new displacement rotor
    returns the new TR and the original length of the T rotor
    """
    R_only = TR(e123)
    T_only = (TR*~R_only).normal()
    t = -2*(T_only|no)
    scale = abs(t)
    t = t/scale
    new_TR = (generate_translation_rotor(t)*R_only).normal()
    return new_TR, scale


def scale_TR_translation(TR, scale):
    """
    Takes in a TR rotor and a scale
    extracts the R and T
    scales the T displacement magnitude by scale
    rebuilds the TR rotor with the new displacement rotor
    returns the new TR rotor
    """
    R_only = TR(e123)
    T_only = (TR*~R_only).normal()
    t = -2*(T_only|no)*scale
    new_TR = (generate_translation_rotor(t)*R_only).normal()
    return new_TR


def left_gmt_generator(mt=layout.gmt):
    # unpack for numba
    k_list, l_list, m_list = mt.coords
    mult_table_vals = mt.data
    gaDims = mt.shape[1]
    val_get_left_gmt_matrix = cf._numba_val_get_left_gmt_matrix

    @numba.njit
    def get_left_gmt(x_val):
        return val_get_left_gmt_matrix(
            x_val, k_list, l_list, m_list, mult_table_vals, gaDims)
    return get_left_gmt


def right_gmt_generator(mt=layout.gmt):
    return left_gmt_generator(mt.T)


get_left_gmt_matrix = left_gmt_generator()
get_right_gmt_matrix = right_gmt_generator()


def get_line_reflection_matrix(lines, n_power=1):
    """
    Generates the matrix that sums the reflection of a point in many lines
    """
    mat2solve = np.zeros((32, 32), dtype=np.float64)
    for Li in lines:
        LiMat = get_left_gmt_matrix(Li.value)
        tmat = np.matmul(np.matmul(LiMat, mask_2minus4), LiMat)
        mat2solve += tmat
    mat = np.matmul(mask1, mat2solve)
    if n_power != 1:
        mat = np.linalg.matrix_power(mat, n_power)
    return mat


@numba.njit
def val_get_line_reflection_matrix(line_array, n_power):
    """
    Generates the matrix that sums the reflection of a point in many lines
    """
    mat2solve = np.zeros((32, 32), dtype=np.float64)
    for i in range(line_array.shape[0]):
        LiMat = get_left_gmt_matrix(line_array[i, :])
        tmat = np.matmul(np.matmul(LiMat, mask_2minus4), LiMat)
        mat2solve += tmat
    mat = np.matmul(mask1, mat2solve)
    if n_power != 1:
        mat = np.linalg.matrix_power(mat, n_power)
    return mat


@numba.njit
def val_truncated_get_line_reflection_matrix(line_array, n_power):
    """
    Generates the truncated matrix that sums the
    reflection of a point in many lines
    n_power should be 1 or a power of 2
    """
    mat2solve = np.zeros((32, 32), dtype=np.float64)
    for i in range(line_array.shape[0]):
        LiMat = get_left_gmt_matrix(line_array[i, :])
        tmat = np.matmul(np.matmul(LiMat, mask_2minus4), LiMat)
        mat2solve += tmat
    mat_val = np.matmul(mask1, mat2solve)
    mat_val = mat_val[1:6, 1:6].copy()/line_array.shape[0]
    if n_power != 1:
        c_pow = 1
        while c_pow < n_power:
            mat_val = np.matmul(mat_val, mat_val)
            c_pow = c_pow*2
    return mat_val


@numba.njit
def val_get_line_intersection(L3_val, Ldd_val):
    """
    Gets the point of intersection of two orthogonal lines that meet
    Xdd = Ldd*no*Ldd + no
    Xddd = L3*Xdd*L3
    Pd = 0.5*(Xdd+Xddd)
    P = -(Pd*ninf*Pd)(1)/(2*(Pd|einf)**2)[0]
    """
    Xdd = gmt_func(gmt_func(Ldd_val, no_val), Ldd_val) + no_val
    Xddd = gmt_func(gmt_func(L3_val, Xdd), L3_val)
    Pd = 0.5*(Xdd+Xddd)
    P = -gmt_func(gmt_func(Pd, ninf_val), Pd)
    imt_value = imt_func(Pd, ninf_val)
    P_denominator = 2*(gmt_func(imt_value, imt_value))[0]
    return project_val(P/P_denominator, 1)


def get_line_intersection(L3, Ldd):
    """
    Gets the point of intersection of two orthogonal lines that meet
    Xdd = Ldd*no*Ldd + no
    Xddd = L3*Xdd*L3
    Pd = 0.5*(Xdd+Xddd)
    P = -(Pd*ninf*Pd)(1)/(2*(Pd|einf)**2)[0]
    """
    return layout.MultiVector(value=val_get_line_intersection(L3.value, Ldd.value))


@numba.njit
def val_midpoint_between_lines(L1_val, L2_val):
    """
    Gets the point that is maximally close to both lines
    Hadfield and Lasenby AGACSE2018
    """
    L3 = val_normalised(L1_val + L2_val)
    Ldd = val_normalised(L1_val - L2_val)
    S = val_normalised(gmt_func(I5_val, val_get_line_intersection(L3, Ldd)))
    return val_normalise_n_minus_1(gmt_func(S, gmt_func(ninf_val, S)))


def midpoint_between_lines(L1, L2):
    """
    Gets the point that is maximally close to both lines
    Hadfield and Lasenby AGACSE2018
    """
    return layout.MultiVector(value=val_midpoint_between_lines(L1.value, L2.value))


def midpoint_of_line_cluster(line_cluster):
    """
    Gets a center point of a line cluster
    Hadfield and Lasenby AGACSE2018
    """
    return layout.MultiVector(value=val_midpoint_of_line_cluster(MVArray(line_cluster).value))


@numba.njit
def val_midpoint_of_line_cluster(array_line_cluster):
    """
    Gets a center point of a line cluster
    Hadfield and Lasenby AGACSE2018
    """
    # Allocate some space for our finished matrix
    accumulator_matrix = np.zeros((32, 32))

    # Loop over our lines and construct the matrix
    for i in range(array_line_cluster.shape[0]):
        # Get the line as a left gmt matrix
        L_i_l = get_left_gmt_matrix(array_line_cluster[i, :])
        # Get the line as a right gmt matrix
        L_i_r = get_right_gmt_matrix(array_line_cluster[i, :])
        # Multiply and add
        accumulator_matrix += L_i_r @ L_i_l

    # Raise the matrix to a very high power
    power_mat = np.linalg.matrix_power(accumulator_matrix / array_line_cluster.shape[0], 256)

    # Get a point that lies on the first line as an approximation to the e-vector
    pp_val = imt_func(array_line_cluster[0, :], eo_val)
    p_start = val_normalise_n_minus_1(project_val(gmt_func(gmt_func(pp_val, ninf_val), pp_val), 1))

    # Apply the matrix
    p_end = project_val(power_mat @ p_start, 1)

    # Remove any junk that has come along with it
    final_point = val_normalise_n_minus_1(project_val(gmt_func(gmt_func(p_end, ninf_val), p_end), 1))
    return final_point


@numba.njit(parallel=NUMBA_PARALLEL)
def val_midpoint_of_line_cluster_grad(array_line_cluster):
    """
    Gets an approximate center point of a line cluster
    Hadfield and Lasenby AGACSE2018
    """
    average_line = val_average_objects(array_line_cluster)
    val_point_track = np.zeros(32)
    for i in range(array_line_cluster.shape[0]):
        p = val_midpoint_between_lines(average_line, array_line_cluster[i, :])
        val_point_track += p
    S = gmt_func(I5_val, val_point_track)
    center_point = val_normalise_n_minus_1(project_val(gmt_func(S, gmt_func(ninf_val, S)), 1))
    # Take a derivative of the cost function at this point
    grad = np.zeros(32)
    for i in range(array_line_cluster.shape[0]):
        l_val = array_line_cluster[i, :]
        grad += (gmt_func(gmt_func(l_val, center_point), l_val))
    grad = val_normalise_n_minus_1(project_val(grad, 1))
    s_val = gmt_func(I5_val, project_val(center_point + grad, 1))
    center_point = val_normalise_n_minus_1(gmt_func(gmt_func(s_val, ninf_val), s_val))
    return center_point


def get_circle_in_euc(circle):
    """ Extracts all the normal stuff for a circle """
    Ic = (circle^ninf).normal()
    GAnormal = get_plane_normal(Ic)
    inPlaneDual = circle*Ic
    mag = float((inPlaneDual|ninf)[0])
    inPlaneDual = -inPlaneDual/mag
    radius_squared = (inPlaneDual*inPlaneDual)[0]
    radius = math.sqrt(abs(radius_squared))
    if radius_squared < 0:
        # We have an imaginary circle, return it as a negative radius as our signal
        radius = -radius
    GAcentre = down(inPlaneDual*(1+0.5*inPlaneDual*ninf))
    return [GAcentre, GAnormal, radius]


def circle_to_sphere(C):
    """
    returns the sphere for which the input circle is the perimeter
    """
    Ic = (C ^ einf).normal()
    sphere = C * Ic * I5
    return sphere


def line_to_point_and_direction(line):
    """
    converts a line to the conformal nearest point to the origin and a
    euc direction vector in direction of the line
    """
    L_star = line*I5
    T = L_star|no
    mhat = -(L_star - T*ninf)*I3
    p = (T^mhat)*I3
    return [p, mhat]


def get_plane_origin_distance(plane):
    """ Get the distance between a given plane and the origin """
    return float(((plane*I5)|no)[0])


def get_plane_normal(plane):
    """ Get the normal to the plane """
    return (plane*I5 - get_plane_origin_distance(plane)*ninf)


def get_nearest_plane_point(plane):
    """ Get the nearest point to the origin on the plane """
    return get_plane_normal(plane)*get_plane_origin_distance(plane)


def disturb_object(mv_object, maximum_translation=0.01, maximum_angle=0.01):
    """ Disturbs an object by a random rotor """
    r = random_rotation_translation_rotor(maximum_translation=maximum_translation, maximum_angle=maximum_angle)
    return (r*mv_object*~r).normal()


def generate_n_clusters(object_generator, n_clusters, n_objects_per_cluster):
    """ Creates n_clusters of random objects """
    object_clusters = []
    for i in range(n_clusters):
        cluster_objects = generate_random_object_cluster(n_objects_per_cluster, object_generator,
                                                         max_cluster_trans=0.5, max_cluster_rot=np.pi / 16)
        object_clusters.append(cluster_objects)
    all_objects = [item for sublist in object_clusters for item in sublist]
    return all_objects, object_clusters


def generate_random_object_cluster(n_objects, object_generator, max_cluster_trans=1.0, max_cluster_rot=np.pi/8):
    """ Creates a cluster of random objects """
    ref_obj = object_generator()
    cluster_objects = []
    for i in range(n_objects):
        r = random_rotation_translation_rotor(maximum_translation=max_cluster_trans, maximum_angle=max_cluster_rot)
        new_obj = apply_rotor(ref_obj, r)
        cluster_objects.append(new_obj)
    return cluster_objects


def random_translation_rotor(maximum_translation=10.0):
    """ generate a random translation rotor """
    return generate_translation_rotor(random_euc_mv(maximum_translation))


def random_rotation_translation_rotor(maximum_translation=10.0, maximum_angle=np.pi):
    """ generate a random combined rotation and translation rotor """
    return (random_translation_rotor(maximum_translation)*random_rotation_rotor(maximum_angle)).normal()


@numba.njit
def project_val(val, grade):
    """ fast grade projection """
    output = np.zeros(32)
    if grade == 0:
        output[0] = val[0]
    elif grade == 1:
        output[1:6] = val[1:6]
    elif grade == 2:
        output[6:16] = val[6:16]
    elif grade == 3:
        output[16:26] = val[16:26]
    elif grade == 4:
        output[26:31] = val[26:31]
    elif grade == 5:
        output[31] = val[31]
    return output


def random_conformal_point(l_max=10):
    """
    Creates a random conformal point
    """
    return up(random_euc_mv(l_max=l_max))


def generate_dilation_rotor(scale):
    """
    Generates a rotor that performs dilation about the origin
    """
    if abs(scale - 1.0) < 0.00001:
        u = np.zeros(32)
        u[0] = 1.0
        return cf.MultiVector(layout, u)
    gamma = math.log(scale)
    return math.cosh(gamma/2) + math.sinh(gamma/2)*(ninf^no)


def generate_translation_rotor(euc_vector_a):
    """
    Generates a rotor that translates objects along the euclidean vector euc_vector_a
    """
    return layout.MultiVector(value=val_generate_translation_rotor(euc_vector_a.value))


@numba.njit
def val_generate_translation_rotor(euc_vector_a):
    """
    Generates a rotor that translates objects along the euclidean vector euc_vector_a
    """
    T = gmt_func(ninf_val, euc_vector_a) / 2
    T[0] += 1
    return T


@numba.njit
def meet_val(a_val, b_val):
    """
    The meet algorithm as described in "A Covariant Approach to Geometry"
    I5*((I5*A) ^ (I5*B))
    """
    return dual_func(omt_func(dual_func(a_val), dual_func(b_val)))


def meet(A, B):
    """
    The meet algorithm as described in "A Covariant Approach to Geometry"
    I5*((I5*A) ^ (I5*B))
    """
    return cf.MultiVector(layout, meet_val(A.value, B.value))


@numba.njit
def val_intersect_line_and_plane_to_point(line_val, plane_val):
    """
    Returns the point at the intersection of a line and plane
    """
    m = meet_val(line_val, plane_val)
    if gmt_func(m, m)[0] < 0.000001:
        return np.array([-1.])
    output = np.zeros(32)
    A = val_normalised(m)
    if A[15] < 0:
        output[1] = A[8]
        output[2] = A[11]
        output[3] = A[14]
    else:
        output[1] = -A[8]
        output[2] = -A[11]
        output[3] = -A[14]
    output = val_up(output)
    return output


def intersect_line_and_plane_to_point(line, plane):
    """
    Returns the point at the intersection of a line and plane
    If there is no intersection it returns None
    """
    ans = val_intersect_line_and_plane_to_point(line.value, plane.value)
    if ans[0] == -1.:
        return None
    return layout.MultiVector(value=ans)


@numba.njit
def val_normalise_n_minus_1(mv_val):
    """
    Normalises a conformal point so that it has an inner product of -1 with einf
    """
    scale = imt_func(mv_val, ninf_val)[0]
    if scale != 0.0:
        return -mv_val/scale
    else:
        raise ZeroDivisionError('Multivector has 0 einf component')


def normalise_n_minus_1(mv):
    """
    Normalises a conformal point so that it has an inner product of -1 with einf
    """
    return layout.MultiVector(value=val_normalise_n_minus_1(mv.value))


def quaternion_and_vector_to_rotor(quaternion, vector):
    """
    Takes in a quaternion and a vector and returns a conformal rotor that
    implements the transformation
    """
    rotation = quaternion_to_rotor(quaternion)(0, 2, 4)
    translation = generate_translation_rotor(vector[0] * e1 + vector[1] * e2 + vector[2] * e3)(0, 2, 4)
    return (translation * rotation).normal()


def get_center_from_sphere(sphere):
    """
    Returns the conformal point at the centre of a sphere by reflecting the
    point at infinity
    """
    center = sphere * ninf * sphere
    return center


def get_radius_from_sphere(sphere):
    """
    Returns the radius of a sphere
    """
    dual_sphere = sphere * I5
    dual_sphere = dual_sphere / (-dual_sphere | ninf)[0]
    return math.sqrt(abs(dual_sphere * dual_sphere))


@numba.njit
def val_point_pair_to_end_points(T):
    """
    Extracts the end points of a point pair bivector
    """
    beta = np.sqrt(abs(gmt_func(T, T)[0]))
    F = T / beta
    P = 0.5*F
    P[0] += 0.5
    P_twiddle = -0.5*F
    P_twiddle[0] += 0.5
    A = val_normalise_n_minus_1(-gmt_func(P_twiddle, imt_func(T, ninf_val)))
    B = val_normalise_n_minus_1(gmt_func(P, imt_func(T, ninf_val)))
    output = np.zeros((2, 32))
    output[0, :] = A
    output[1, :] = B
    return output


def point_pair_to_end_points(T):
    """
    Extracts the end points of a point pair bivector
    """
    output = val_point_pair_to_end_points(T.value)
    A = layout.MultiVector(value=output[0, :])
    B = layout.MultiVector(value=output[1, :])
    return A, B


def euc_dist(conf_mv_a, conf_mv_b):
    """ Returns the distance between two conformal points """
    dot_result = (conf_mv_a|conf_mv_b)[0]
    if dot_result < 0.0:
        return math.sqrt(-2.0*dot_result)
    else:
        return 0.0


@numba.jit
def dorst_norm_val(sigma_val):
    """ Square Root of Rotors - Implements the norm of a rotor"""
    sigma_4 = project_val(sigma_val, 4)
    sqrd_ans = sigma_val[0] ** 2 - gmt_func(sigma_4, sigma_4)[0]
    return math.sqrt(sqrd_ans)


@numba.njit
def check_sigma_for_positive_root_val(sigma_val):
    """ Square Root of Rotors - Checks for a positive root """
    return (sigma_val[0] + dorst_norm_val(sigma_val)) > 0


def check_sigma_for_positive_root(sigma):
    """ Square Root of Rotors - Checks for a positive root """
    return check_sigma_for_positive_root_val(sigma.value)


@numba.njit
def check_sigma_for_negative_root_val(sigma_value):
    """ Square Root of Rotors - Checks for a negative root """
    return (sigma_value[0] - dorst_norm_val(sigma_value)) > 0


def check_sigma_for_negative_root(sigma):
    """ Square Root of Rotors - Checks for a negative root """
    return check_sigma_for_negative_root_val(sigma.value)


@numba.njit
def check_infinite_roots_val(sigma_value):
    """ Square Root of Rotors - Checks for a infinite roots """
    return (sigma_value[0] + dorst_norm_val(sigma_value)) < 0.0000000001


def check_infinite_roots(sigma):
    """ Square Root of Rotors - Checks for a infinite roots """
    return check_infinite_roots_val(sigma.value)


@numba.njit
def positive_root_val(sigma_val):
    """
    Square Root of Rotors - Evaluates the positive root
    Square Root of Rotors - Evaluates the positive root
    """
    norm_s = dorst_norm_val(sigma_val)
    denominator = (math.sqrt(2) * math.sqrt(sigma_val[0] + norm_s))
    result = sigma_val/denominator
    result[0] += norm_s/denominator
    return result


@numba.njit
def negative_root_val(sigma_val):
    """
    Square Root of Rotors - Evaluates the positive root
    """
    norm_s = dorst_norm_val(sigma_val)
    denominator = (math.sqrt(2) * math.sqrt(sigma_val[0] - norm_s))
    result = sigma_val/denominator
    result[0] -= norm_s/denominator
    return result


def positive_root(sigma):
    """
    Square Root of Rotors - Evaluates the positive root
    """
    res_val = positive_root_val(sigma.value)
    return cf.MultiVector(layout, res_val)


def negative_root(sigma):
    """ Square Root of Rotors - Evaluates the negative root """
    res_val = negative_root_val(sigma.value)
    return cf.MultiVector(layout, res_val)


@numba.njit
def general_root_val(sigma_value):
    """
    Square Root and Logarithm of Rotors
    in 3D Conformal Geometric Algebra
    Using Polar Decomposition
    Leo Dorst and Robert Valkenburg
    """
    if check_sigma_for_positive_root_val(sigma_value):
        k = positive_root_val(sigma_value)
        output = np.zeros((2, sigma_value.shape[0]))
        output[0, :] = k.copy()
        return output
    elif check_sigma_for_negative_root_val(sigma_value):
        k = positive_root_val(sigma_value)
        k2 = negative_root_val(sigma_value)
        output = np.zeros((2, sigma_value.shape[0]))
        output[0, :] = k.copy()
        output[1, :] = k2.copy()
        return output
    elif check_infinite_roots_val(sigma_value):
        output = np.zeros((2, sigma_value.shape[0]))
        output[:, 0] = 1.0
        return output
    else:
        raise ValueError('No root exists')


def general_root(sigma):
    """ The general case of the root of a grade 0, 4 multivector """
    output = general_root_val(sigma.value)
    return [cf.MultiVector(layout, output[0, :].copy()), cf.MultiVector(layout, output[1, :].copy())]


@numba.njit
def val_annihilate_k(K_val, C_val):
    """ Removes K from C = KX via (K[0] - K[4])*C """
    k_4 = -project_val(K_val, 4)
    k_4[0] += K_val[0]
    return val_normalised(gmt_func(k_4, C_val))


def annihilate_k(K, C):
    """ Removes K from C = KX via (K[0] - K[4])*C """
    return cf.MultiVector(layout, val_annihilate_k(K.value, C.value))


@numba.njit
def pos_twiddle_root_val(C_value):
    """
    Square Root and Logarithm of Rotors
    in 3D Conformal Geometric Algebra
    Using Polar Decomposition
    Leo Dorst and Robert Valkenburg
    """
    sigma_val = gmt_func(C_value, adjoint_func(C_value))
    k_value = general_root_val(sigma_val)
    output = np.zeros((2, 32))
    output[0, :] = val_annihilate_k(k_value[0, :], C_value)
    output[1, :] = val_annihilate_k(k_value[1, :], C_value)
    return output


@numba.njit
def neg_twiddle_root_val(C_value):
    """
    Square Root and Logarithm of Rotors
    in 3D Conformal Geometric Algebra
    Using Polar Decomposition
    Leo Dorst and Robert Valkenburg
    """
    sigma_val = -gmt_func(C_value, adjoint_func(C_value))
    k_value = general_root_val(sigma_val)
    output = np.zeros((2, 32))
    output[0, :] = val_annihilate_k(k_value[0, :], C_value)
    output[1, :] = val_annihilate_k(k_value[1, :], C_value)
    return output


def pos_twiddle_root(C):
    """
    Square Root and Logarithm of Rotors
    in 3D Conformal Geometric Algebra
    Using Polar Decomposition
    Leo Dorst and Robert Valkenburg
    """
    output = pos_twiddle_root_val(C.value)
    return [cf.MultiVector(layout, output[0, :]), cf.MultiVector(layout, output[1, :])]


def neg_twiddle_root(C):
    """
    Square Root and Logarithm of Rotors
    in 3D Conformal Geometric Algebra
    Using Polar Decomposition
    Leo Dorst and Robert Valkenburg
    """
    output = neg_twiddle_root_val(C.value)
    return [cf.MultiVector(layout, output[0, :]), cf.MultiVector(layout, output[1, :])]


def square_roots_of_rotor(R):
    """
    Square Root and Logarithm of Rotors
    in 3D Conformal Geometric Algebra
    Using Polar Decomposition
    Leo Dorst and Robert Valkenburg
    """
    return pos_twiddle_root(1 + R)


def n_th_rotor_root(R, n):
    """
    Takes the n_th root of rotor R
    n must be a power of 2
    """
    if not (((n & (n - 1)) == 0) and n != 0):
        raise ValueError('n is not a power of 2')
    if n == 1:
        return R
    else:
        return n_th_rotor_root(square_roots_of_rotor(R)[0], int(n/2))


def interp_objects_root(C1, C2, alpha):
    """
    Hadfield and Lasenby, Direct Linear Interpolation of Geometric Objects, AGACSE2018
    Directly linearly interpolates conformal objects
    Return a valid object from the addition result C
    """
    C = (1 - alpha) * C1 + alpha*C2
    C3 = normalised(neg_twiddle_root(C)[0])
    if cf.grade_obj(C1, 0.00001) != cf.grade_obj(C3, 0.00001):
        raise ValueError('Created object is not same grade')
    return C3


def general_object_interpolation(object_alpha_array, object_list, new_alpha_array, kind='linear'):
    """
    Hadfield and Lasenby, Direct Linear Interpolation of Geometric Objects, AGACSE2018
    This is a general interpolation through the
    """
    obj_array = np.transpose(MVArray(object_list).value)
    f = interp1d(object_alpha_array, obj_array, kind=kind)
    new_value_array = np.transpose(f(new_alpha_array))
    new_conf_array = MVArray.from_value_array(layout, new_value_array)
    return [normalised(neg_twiddle_root(C)[0]) for C in new_conf_array]


@numba.njit
def val_average_objects_with_weights(obj_array, weights_array):
    """
    Hadfield and Lasenby, Direct Linear Interpolation of Geometric Objects, AGACSE2018
    Directly averages conformal objects
    Return a valid object from the addition result C
    """
    C_val = np.zeros(32)
    for i in range(obj_array.shape[0]):
        C_val += obj_array[i, :]*weights_array[i]
    C3 = val_normalised(neg_twiddle_root_val(C_val)[0, :])
    return C3


@numba.njit
def val_average_objects(obj_array):
    """
    Hadfield and Lasenby, Direct Linear Interpolation of Geometric Objects, AGACSE2018
    Directly averages conformal objects
    Return a valid object from the addition result C
    """
    C_val = np.zeros(32)
    for i in range(obj_array.shape[0]):
        C_val += obj_array[i, :]
    C_val = C_val / obj_array.shape[0]
    C3 = val_normalised(neg_twiddle_root_val(C_val)[0, :])
    return C3


def average_objects(obj_list, weights=[], check_grades=True):
    """
    Hadfield and Lasenby, Direct Linear Interpolation of Geometric Objects, AGACSE2018
    Directly averages conformal objects
    Return a valid object from the addition result C
    """
    if len(weights) == len(obj_list):
        C = sum([o * w for o, w in zip(obj_list, weights)])
    else:
        C = sum(obj_list) / len(obj_list)
    C3 = normalised(neg_twiddle_root(C)[0])
    if check_grades:
        if cf.grade_obj(obj_list[0], 0.00001) != cf.grade_obj(C3, 0.00001):
            raise ValueError('Created object is not same grade \n' + str(obj_list[0]) + '\n' + str(C3))
    return C3


def rotor_between_objects(X1, X2):
    """
    Lasenby and Hadfield AGACSE2018
    For any two conformal objects X1 and X2 this returns a rotor that takes X1 to X2
    Return a valid object from the addition result 1 + gamma*X2X1
    """
    return cf.MultiVector(
        layout,
        val_rotor_between_objects_root(val_normalised(X1.value),
                                       val_normalised(X2.value))
    )


def TRS_between_rounds(X1, X2):
    """
    Calculate the TRS rotor between any pair of rounds of the same grade
    Bring rounds to origin, line up carriers, calculate scale
    """
    T1 = generate_translation_rotor(-down((X1 * einf * X1)(1)))
    X1h = normalised(T1 * X1 * ~T1)
    T2 = generate_translation_rotor(-down((X2 * einf * X2)(1)))
    X2h = normalised(T2 * X2 * ~T2)
    X1f = normalised(X1h ^ einf)
    X2f = normalised(X2h ^ einf)
    Rc = rotor_between_objects(X1f, X2f)
    S = generate_dilation_rotor(get_radius_from_sphere(normalised(X2h*X2f*I5))/get_radius_from_sphere(normalised(X1h*X1f*I5)))
    return normalised((~T2)*S*Rc*T1)


def motor_between_rounds(X1, X2):
    """
    Calculate the motor between any pair of rounds of the same grade
    Line up the carriers, then line up the centers
    """
    return layout.MultiVector(value=val_motor_between_rounds(X1.value, X2.value))


@numba.njit
def val_motor_between_rounds(X1, X2):
    """
    Calculate the motor between any pair of rounds of the same grade
    Line up the carriers, then line up the centers

    Optimised form of this:

    R = rotor_between_objects(normalised(X1^einf), normalised(X2^einf))
    X3 = apply_rotor(X1, R)
    C1 = normalise_n_minus_1((X3 * einf * X3)(1)).value[1:4]
    C2 = normalise_n_minus_1((X2 * einf * X2)(1)).value[1:4]
    t = layout.MultiVector()
    t.value[1:4] = C2 - C1
    T = generate_translation_rotor(t)
    return normalised(T*R)
    """
    F1 = val_normalised(omt_func(X1, ninf_val))
    F2 = val_normalised(omt_func(X2, ninf_val))

    if np.abs(F1[31]) > 1E-5:
        # Its spheres we are dealing with
        R = unit_scalar_mv_val
        X3 = X1
    else:
        R = val_rotor_between_objects_root(F1, F2)
        X3 = val_apply_rotor(X1, R)

    C1 = val_normalise_n_minus_1(project_val(gmt_func(gmt_func(X3, ninf_val), X3), 1))
    C2 = val_normalise_n_minus_1(project_val(gmt_func(gmt_func(X2, ninf_val), X2), 1))

    t = np.zeros(32)
    t[1:4] = (C2 - C1)[1:4]
    T = val_generate_translation_rotor(t)
    return val_normalised(gmt_func(T, R))


@numba.njit
def val_motor_between_objects(X1, X2):
    """
    Calculates a motor that takes X1 to X2
    """
    carrier = omt_func(X1, ninf_val)
    if np.sum(np.abs(carrier)) < 1E-4:
        # They are flats
        return val_rotor_between_objects_root(X1, X2)
    else:
        # Rounds
        return val_motor_between_rounds(X1, X2)


def motor_between_objects(X1, X2):
    """
    Calculates a motor that takes X1 to X2
    """
    return layout.MultiVector(value=val_motor_between_objects(X1.value, X2.value))


def calculate_S_over_mu(X1, X2):
    """
    Lasenby and Hadfield AGACSE2018
    For any two conformal objects X1 and X2 this returns a factor that corrects
    the X1 + X2 back to a blade
    """
    gamma1 = (X1 * X1)[0]
    gamma2 = (X2 * X2)[0]

    M12 = X1 * X2 + X2 * X1
    K = 2 + gamma1 * M12
    mu = (K[0]**2 - K(4)**2)[0]

    if sum(np.abs(M12(4).value)) > 0.0000001:
        lamb = (-(K(4) * K(4)))[0]
        mu = K[0] ** 2 + lamb
        root_mu = np.sqrt(mu)
        if abs(lamb) < 0.0000001:
            beta = 1.0 / (2 * np.sqrt(K[0]))
        else:
            beta_sqrd = 1 / (2 * (root_mu + K[0]))
            beta = np.sqrt(beta_sqrd)
        S = -gamma1/(2*beta) + beta*M12(4)
        return S/np.sqrt(mu)
    else:
        S = np.sqrt(abs(K[0]))
    return S/np.sqrt(mu)


I5eoval = (I5 * eo).value
biv3dmask = (e12+e13+e23).value


@numba.njit
def val_rotor_between_objects_root(X1, X2):
    """
    Lasenby and Hadfield AGACSE2018
    For any two conformal objects X1 and X2 this returns a rotor that takes X1 to X2
    Uses the square root of rotors for efficiency and numerical stability
    """
    X21 = gmt_func(X2, X1)
    X12 = gmt_func(X1, X2)
    gamma = gmt_func(X1, X1)[0]
    if gamma > 0:
        C_val = gamma*gmt_func(X2, X1)
        C_val[0] += 1
        if abs(C_val[0]) < 1E-6:
            R = val_normalised(project_val(gmt_func(I5eoval, X21), 2))
            return val_normalised(gmt_func(R, val_rotor_between_objects_root(X1, -X2)))
        return val_normalised(pos_twiddle_root_val(C_val)[0, :])
    else:
        C_val = -X21
        C_val[0] += 1
        if abs(C_val[0]) < 1E-6:
            R = project_val(gmt_func(I5eoval, X21), 2)
            R = val_normalised(project_val(gmt_func(R, biv3dmask), 2))
            R2 = val_normalised(val_rotor_between_objects_root(val_apply_rotor(X1, R), X2))
            return val_normalised(gmt_func(R2, R))
        else:
            return val_normalised(C_val)


@numba.njit
def val_rotor_between_objects_explicit(X1, X2):
    """
    Lasenby and Hadfield AGACSE2018
    For any two conformal objects X1 and X2 this returns a rotor that takes X1 to X2

    Implements an optimised version of::

        gamma1 = (X1 * X1)[0]
        gamma2 = (X2 * X2)[0]

        M12 = X1*X2 + X2*X1
        K = 2 + gamma1*M12

        if np.sum(np.abs(K.value)) < 0.0000001:
            return 1 + 0*e1

        if sum(np.abs(M12(4).value)) > 0.0000001:
            lamb = (-(K(4) * K(4)))[0]
            mu = K[0]**2 + lamb
            root_mu = np.sqrt(mu)
            if abs(lamb) < 0.0000001:
                beta = 1.0/(2*np.sqrt(K[0]))
            else:
                beta_sqrd = 1/(2*(root_mu + K[0]))
                beta = np.sqrt(beta_sqrd)
            R = ( beta*K(4) - (1/(2*beta)) )*(1 + gamma1*X2*X1)/(root_mu)
            return R
        else:
            return (1 + gamma1*X2*X1)/(np.sqrt(abs(K[0])))

    """
    gamma1 = gmt_func(X1, X1)[0]
    gamma2 = gmt_func(X2, X2)[0]

    # if abs(abs(gamma1) - 1.0) > 0.00001:
    #     raise ValueError('X1 must be normalised to give abs(X1*X1) == 1')
    # elif abs(abs(gamma2) - 1.0) > 0.00001:
    #     raise ValueError('X1 must be normalised to give abs(X1*X1) == 1')
    # elif abs(gamma1 - gamma2) > 0.00001:
    #     raise ValueError('X1 and X2 must square to the same value')

    M12_val = gmt_func(X1, X2) + gmt_func(X2, X1)
    K_val = gamma1 * M12_val
    K_val[0] = K_val[0] + 2

    if np.sum(np.abs(K_val)) < 0.0000001:
        return unit_scalar_mv_val

    if np.sum(np.abs(project_val(M12_val, 4))) > 0.00001:
        K_val_4 = project_val(K_val, 4)
        lamb = -gmt_func(K_val_4, K_val_4)[0]
        mu = K_val[0] ** 2 + lamb

        root_mu = np.sqrt(mu)
        beta_sqrd = 1 / (2*root_mu + 2*K_val[0])
        beta = np.sqrt(beta_sqrd)

        temp_1 = beta * K_val_4
        temp_1[0] = temp_1[0] - (1 / (2 * beta))

        temp_2 = gamma1*gmt_func(X2, X1)
        temp_2[0] = temp_2[0] + 1.0

        R_val = gmt_func(temp_1, temp_2) / root_mu
        return R_val
    else:
        temp_1 = gamma1 * gmt_func(X2, X1)
        temp_1[0] += 1.0
        return temp_1/(np.sqrt(abs(K_val[0])))


sparse_line_gmt = layout.gmt_func_generator(grades_a=[3], grades_b=[3])


@numba.njit
def val_norm(mv_val):
    """ Returns sqrt(abs(~A*A)) """
    return np.sqrt(np.abs(gmt_func(adjoint_func(mv_val), mv_val)[0]))


def norm(mv):
    """ Returns sqrt(abs(~A*A)) """
    return val_norm(mv.value)


@numba.njit
def val_normalised(mv_val):
    """ Returns A/sqrt(abs(~A*A)) """
    return mv_val/val_norm(mv_val)


def normalised(mv):
    """ fast version of the normal() function """
    return cf.MultiVector(layout, val_normalised(mv.value))


@numba.njit
def val_rotor_between_lines(L1_val, L2_val):
    """ Implements a very optimised rotor line to line extraction """
    L21_val = sparse_line_gmt(L2_val, L1_val)
    L12_val = sparse_line_gmt(L1_val, L2_val)
    K_val = L21_val + L12_val
    K_val[0] += 2.0
    beta_val = project_val(K_val, 4)
    alpha = 2 * K_val[0]

    denominator = np.sqrt(alpha / 2)
    numerator_val = -beta_val/alpha
    numerator_val[0] += 1.0
    normalisation_val = numerator_val/denominator

    output_val = L21_val
    output_val[0] += 1
    return gmt_func(normalisation_val, output_val)


def rotor_between_lines(L1, L2):
    """ return the rotor between two lines """
    return cf.MultiVector(layout, val_rotor_between_lines(L1.value, L2.value))


def rotor_between_planes(P1, P2):
    """ return the rotor between two planes """
    return cf.MultiVector(layout, val_rotor_rotor_between_planes(P1.value, P2.value))


@numba.njit
def val_rotor_rotor_between_planes(P1_val, P2_val):
    """ return the rotor between two planes """
    P21_val = -gmt_func(P2_val, P1_val)
    P21_val[0] += 1
    return val_normalised(P21_val)


def random_bivector():
    r"""
    Creates a random bivector on the form described by R. Wareham in
    Mesh Vertex Pose and Position Interpolation using Geometric Algebra.
    $$ B =  ab + c*n_{\inf}$$ where $a, b, c \in \mathcal(R)^3$
    """
    a = random_euc_mv()
    c = random_euc_mv()
    return a * I3 + c * ninf


def standard_point_pair_at_origin():
    """ Creates a standard point pair at the origin """
    return (up(-0.5*e1)^up(0.5*e1)).normal()


def random_point_pair_at_origin():
    """
    Creates a random point pair bivector object at the origin
    """
    mv_a = random_euc_mv()
    plane_a = (mv_a*I3).normal()

    mv_b = plane_a*mv_a*plane_a
    pp = (up(mv_a) ^ up(mv_b)).normal()
    return pp


def random_point_pair():
    """
    Creates a random point pair bivector object
    """
    mv_a = random_euc_mv()
    mv_b = random_euc_mv()
    pp = (up(mv_a) ^ up(mv_b)).normal()
    return pp


def standard_line_at_origin():
    """ Creates a standard line at the origin """
    return (standard_point_pair_at_origin()^einf).normal()


def random_line_at_origin():
    """
    Creates a random line at the origin
    """
    pp = (random_point_pair_at_origin()^einf).normal()
    return pp


def random_line():
    """
    Creates a random line
    """
    mv_a = random_euc_mv()
    mv_b = random_euc_mv()
    line_a = ((up(mv_a) ^ up(mv_b) ^ ninf)).normal()
    return line_a


def random_circle_at_origin():
    """
    Creates a random circle at the origin
    """
    mv_a = random_euc_mv()
    mv_r = random_euc_mv()
    r = generate_rotation_rotor(np.pi/2, mv_a, mv_r)
    mv_b = r*mv_a*~r
    mv_c = r * mv_b * ~r
    pp = (up(mv_a) ^ up(mv_b) ^ up(mv_c)).normal()
    return pp


def random_circle():
    """
    Creates a random circle
    """
    mv_a = val_random_euc_mv()
    mv_b = val_random_euc_mv()
    mv_c = val_random_euc_mv()
    return layout.MultiVector(value=val_normalised(omt_func(omt_func(val_up(mv_a), val_up(mv_b)), val_up(mv_c))))


def random_sphere_at_origin():
    """
    Creates a random sphere at the origin
    """
    C = random_circle_at_origin()
    sphere = circle_to_sphere(C)
    return sphere


def random_sphere():
    """
    Creates a random sphere
    """
    mv_a = random_euc_mv()
    mv_b = random_euc_mv()
    mv_c = random_euc_mv()
    mv_d = random_euc_mv()
    sphere = ((up(mv_a) ^ up(mv_b) ^ up(mv_c) ^ up(mv_d))).normal()
    return sphere


def random_plane_at_origin():
    """
    Creates a random plane at the origin
    """
    c = random_circle_at_origin()
    plane = (c ^ einf).normal()
    return plane


def random_plane():
    """
    Creates a random plane
    """
    c = random_circle()
    plane = (c ^ ninf).normal()
    return plane


@numba.njit
def val_apply_rotor(mv_val, rotor_val):
    """ Applies rotor to multivector in a fast way - JITTED """
    return gmt_func(rotor_val, gmt_func(mv_val, adjoint_func(rotor_val)))


def apply_rotor(mv_in, rotor):
    """ Applies rotor to multivector in a fast way """
    return cf.MultiVector(layout, val_apply_rotor(mv_in.value, rotor.value))


@numba.njit
def val_apply_rotor_inv(mv_val, rotor_val, rotor_val_inv):
    """ Applies rotor to multivector in a fast way takes pre computed adjoint"""
    return gmt_func(rotor_val, gmt_func(mv_val, rotor_val_inv))


def apply_rotor_inv(mv_in, rotor, rotor_inv):
    """ Applies rotor to multivector in a fast way takes pre computed adjoint"""
    return cf.MultiVector(layout, val_apply_rotor_inv(mv_in.value, rotor.value, rotor_inv.value))


@numba.njit
def mult_with_ninf(mv):
    """ Convenience function for multiplication with ninf """
    return gmt_func(mv, ninf_val)


# @numba.njit
def val_convert_2D_polar_line_to_conformal_line(rho, theta):
    """ Converts a 2D polar line to a conformal line """
    a = np.cos(theta)
    b = np.sin(theta)
    x0 = a * rho
    y0 = b * rho
    x1 = int(x0 + 10000 * (-b))
    y1 = int(y0 + 10000 * (a))
    x2 = int(x0 - 10000 * (-b))
    y2 = int(y0 - 10000 * (a))
    p1_val = val_convert_2D_point_to_conformal(x1, y1)
    p2_val = val_convert_2D_point_to_conformal(x2, y2)
    line_val = omt_func(omt_func(p1_val, p2_val), ninf_val)
    line_val = line_val/abs(cf.MultiVector(layout, line_val))
    return line_val


def convert_2D_polar_line_to_conformal_line(rho, theta):
    """ Converts a 2D polar line to a conformal line """
    line_val = val_convert_2D_polar_line_to_conformal_line(rho, theta)
    return cf.MultiVector(layout, line_val)


@numba.njit
def val_up(mv_val):
    """ Fast jitted up mapping """
    temp = np.zeros(32)
    temp[0] = 0.5
    return mv_val - no_val + omt_func(temp, gmt_func(gmt_func(mv_val, mv_val), ninf_val))


def fast_up(mv):
    """ Fast up mapping """
    return cf.MultiVector(layout, val_up(mv.value))


@numba.njit
def val_normalInv(mv_val):
    """ A fast, jitted version of normalInv """
    Madjoint_val = adjoint_func(mv_val)
    MadjointM = gmt_func(Madjoint_val, mv_val)[0]
    return Madjoint_val / MadjointM


@numba.njit
def val_homo(mv_val):
    """ A fast, jitted version of homo() """
    return gmt_func(mv_val, val_normalInv(imt_func(-mv_val, ninf_val)))


@numba.njit
def val_down(mv_val):
    """ A fast, jitted version of down() """
    return gmt_func(omt_func(val_homo(mv_val), E0_val), E0_val)


def fast_down(mv):
    """ A fast version of down() """
    return cf.MultiVector(layout, val_down(mv.value))


def val_distance_point_to_line(point, line):
    """
    Returns the euclidean distance between a point and a line
    """
    return float(abs(cf.MultiVector(layout, omt_func(point, line))))


@numba.njit
def val_convert_2D_point_to_conformal(x, y):
    """ Convert a 2D point to conformal """
    mv_val = np.zeros(32)
    mv_val[1] = x
    mv_val[2] = y
    return val_up(mv_val)


def convert_2D_point_to_conformal(x, y):
    """ Convert a 2D point to conformal """
    return cf.MultiVector(layout, val_convert_2D_point_to_conformal(x, y))


def distance_polar_line_to_euc_point_2d(rho, theta, x, y):
    """ Return the distance between a polar line and a euclidean point in 2D """
    point = val_convert_2D_point_to_conformal(x, y)
    line = val_convert_2D_polar_line_to_conformal_line(rho, theta)
    return val_distance_point_to_line(point, line)


dual_gmt_func = layout.gmt_func_generator(grades_a=[5], grades_b=[0, 1, 2, 3, 4, 5])


@numba.njit
def dual_func(a_val):
    """
    Fast dual
    """
    return dual_gmt_func(I5_val, a_val)


def fast_dual(a):
    """
    Fast dual
    """
    return cf.MultiVector(layout, dual_func(a.value))


class ConformalMVArray(cf.MVArray):
    """
    This class is for storing arrays of conformal multivectors
    """
    def draw(self):
        '''
        display mvarray using a given visualization backend

        currently supports pyganja.
        '''
        if pyganja_available:
            return draw([mv for mv in self])
        else:
            pass

    def up(self):
        """
        Up mapping
        """
        return v_up(self)

    def down(self):
        """
        Down mapping
        """
        return v_down(self)

    def dual(self):
        """
        Dualisation
        """
        return v_dual(self)

    def apply_rotor(self, R):
        """
        Application of a rotor
        """
        R_inv = ~R
        return v_apply_rotor_inv(self, R, R_inv)

    def apply_rotor_inv(self, R, R_inv):
        """
        Application of a rotor with precomputed inverse
        """
        return v_apply_rotor_inv(self, R, R_inv)

    @property
    def value(self):
        """
        Return an np array of the values of multivectors
        """
        return np.array([mv.value for mv in self])

    @staticmethod
    def from_value_array(value_array):
        """
        Constructs an array of mvs from a value array
        """
        return ConformalMVArray(v_new_mv(value_array))


v_dual = np.vectorize(fast_dual, otypes=[ConformalMVArray])
v_new_mv = np.vectorize(lambda v: cf.MultiVector(layout, v), otypes=[ConformalMVArray], signature='(n)->()')
v_up = np.vectorize(fast_up, otypes=[ConformalMVArray])
v_down = np.vectorize(fast_down, otypes=[ConformalMVArray])
v_apply_rotor_inv = np.vectorize(apply_rotor_inv, otypes=[ConformalMVArray])
v_meet = np.vectorize(meet, otypes=[ConformalMVArray], signature='(),()->()')




from .object_clustering import assign_measurements_to_objects_matrix, compare_labels
from .rotor_estimation import estimate_rotor_objects, estimate_rotor_objects_subsample, \
    sequential_object_rotor_estimation, \
    estimate_rotor_objects_subsample_sequential
from .rotor_parameterisation import interpolate_TR_rotors
from .cost_functions import val_rotor_cost_sparse
from . import apply_rotor
from .cost_functions import object_set_cost_matrix, object_cost_function, check_p_cost
import numpy as np
from clifford.g3c import *
from clifford.tools.g3c.cuda import sequential_rotor_estimation_cuda_mvs

import clifford as cf


def fingerprint_function(obj, other_objects, nbins=10, max_cost=100.0, cost_func=check_p_cost):
    """
    Bins the cost of the object to all the other objects to make a metric
    for the scene
    """
    cost_list = [cost_func(obj, o_obj) for o_obj in other_objects]
    counts, bins = np.histogram(cost_list, bins=nbins, range=(0.0, max_cost))
    return counts


def match_by_fingerprint(reference_model, query_model, nbins=None, max_cost=None, cost_func=check_p_cost):
    """
    Matches the objects in two scenes based on their fingerprint functions
    """
    if nbins is None:
        nbins = len(query_model)
    if max_cost is None:
        a = np.max(np.max(np.abs(np.array([[cost_func(obj, o_obj) for o_obj in query_model] for obj in query_model]))))
        b = np.max(np.max(np.abs(np.array([[cost_func(obj, o_obj) for o_obj in reference_model] for obj in reference_model]))))
        max_cost = max(a, b)
    labels = []
    min_costs = []
    for j in range(len(query_model)):
        query_bins = fingerprint_function(query_model[j], query_model,
                                          nbins=nbins, max_cost=max_cost, cost_func=cost_func)

        min_cost_match = np.inf
        min_cost_ind = 0
        for i in range(len(reference_model)):
            ref_plane_bins = fingerprint_function(reference_model[i], reference_model,
                                                  nbins=nbins, max_cost=max_cost, cost_func=cost_func)
            cost_match = np.sum(np.abs(query_bins-ref_plane_bins))
            if cost_match < min_cost_match:
                min_cost_match = cost_match
                min_cost_ind = i
        labels.append(min_cost_ind)
        min_costs.append(min_cost_match)
    return labels, min_costs


def iterative_model_match_sequential(reference_model, query_model, iterations=100,
                                     object_type='generic', cuda=False, print_rotor=False,
                                     r_track=None, start_labels=None):
    """
    Matches the query model to the reference model and estimates the motor between them
    Assumes that every query model item has a corresponding reference model item, multiple
    query model items can match the same reference model item. Uses the sequential rotor estimation
    """

    # Get the starting labels
    if start_labels is None:
        labels, costs = assign_measurements_to_objects_matrix(reference_model, query_model,
                                                              object_type=object_type, cuda=cuda)
    else:
        labels = [+l for l in start_labels]
    old_labels = [l for l in labels]
    remapped_objects = [o for o in query_model]
    r_est = 1.0 + 0.0*e1
    assert iterations > 0, 'Must have at least 1 iteration'
    for i in range(iterations):
        # Reorder
        reordered_list_ref = [reference_model[i] for i in labels]
        # Estimate the rotor
        r_est_update, exit_flag = sequential_object_rotor_estimation(reordered_list_ref, remapped_objects,
                                                                     random_sequence=True, n_iterations=10,
                                                                     object_type=object_type)
        # Now update our running estimate
        r_est = (r_est_update*r_est)
        r_est = r_est.normal()
        # Re map with our new rotor
        remapped_objects = [apply_rotor(l, r_est).normal() for l in query_model]
        # Get the new matching
        labels, costs = assign_measurements_to_objects_matrix(reference_model, remapped_objects,
                                                              object_type=object_type, cuda=cuda)
        if r_track is not None:
            r_track.append(r_est)
        if print_rotor:
            print(r_est)
        print(i)
        if compare_labels(old_labels, labels):
            return labels, costs, r_est
        old_labels = [l for l in labels]
    return labels, costs, r_est


def iterative_model_match(reference_model, query_model, iterations=100,
                          object_type='generic', cuda=False, start_labels=None,
                          symmetric=False):
    """
    Matches the query model to the reference model and estimates the motor between them
    Assumes that every query model item has a corresponding reference model item, multiple
    query model items can match the same reference model item
    """

    # Get the starting labels
    if start_labels is None:
        labels, costs = assign_measurements_to_objects_matrix(reference_model, query_model,
                                                              object_type=object_type, cuda=cuda,
                                                              symmetric=symmetric)
    else:
        labels = [+l for l in start_labels]
    old_labels = [+l for l in labels]
    remapped_objects = [o for o in query_model]
    r_est = 1.0 + 0.0*e1
    assert iterations > 0, 'Must have at least 1 iteration'
    for i in range(iterations):
        # Reorder
        reordered_list_ref = [reference_model[i] for i in labels]
        # Estimate the rotor
        r_est_update, cost = estimate_rotor_objects(reordered_list_ref, remapped_objects,
                                                    object_type=object_type,
                                                    symmetric=symmetric)
        # Now update our running estimate
        r_est = (r_est_update*r_est)
        r_est = r_est.normal()

        # Re map with our new rotor
        remapped_objects = [apply_rotor(l, r_est).normal() for l in query_model]
        # Get the new matching
        labels, costs = assign_measurements_to_objects_matrix(reference_model, remapped_objects,
                                                              object_type=object_type, cuda=cuda,
                                                              symmetric=symmetric)
        if compare_labels(old_labels, labels):
            return labels, costs, r_est
        old_labels = [+l for l in labels]
        print(i)
    return labels, costs, r_est


def REFORM(reference_model, query_model, n_samples=100, objects_per_sample=5,
           iterations=100, covergence_threshold=0.00000001,
           pool_size=1, object_type='generic', cuda=False,
           print_rotor=False, start_labels=None, motor=True):
    #  Get the starting labels
    if start_labels is None:
        labels, costs = assign_measurements_to_objects_matrix(reference_model, query_model,
                                                              object_type=object_type, cuda=cuda)
    else:
        labels = [+l for l in start_labels]

    min_global_cost = np.inf
    min_global_rotor = 1.0 + 0.0 * e1

    r_est = 1.0 + 0.0 * e1
    remapped_objects = [o for o in query_model]

    assert iterations > 0, 'Must have at least 1 iteration'
    for i in range(iterations):
        # Reorder and estimate the rotor
        reordered_list_a = [reference_model[i] for i in labels]
        r_est_update, cost = estimate_rotor_objects_subsample(reordered_list_a, remapped_objects,
                                                              n_samples,
                                                              objects_per_sample,
                                                              pool_size=pool_size,
                                                              object_type=object_type,
                                                              motor=motor)
        r_est = (r_est_update * r_est)
        r_est = r_est.normal()
        # Re map with our new rotor
        remapped_objects = [apply_rotor(l, r_est).normal() for l in query_model]
        # Get the new matching
        labels, costs = assign_measurements_to_objects_matrix(reference_model, remapped_objects,
                                                              object_type=object_type, cuda=cuda)
        current_cost = np.sum(costs)
        if print_rotor:
            print(r_est)
        print(i, current_cost, covergence_threshold)
        if current_cost < min_global_cost:
            min_global_cost = current_cost
            min_global_rotor = +r_est
        if current_cost < covergence_threshold:
            return labels, costs, r_est
    # Re map with our new rotor
    remapped_objects = [apply_rotor(l, min_global_rotor).normal() for l in query_model]
    # Get the new matching
    labels, costs = assign_measurements_to_objects_matrix(reference_model, remapped_objects, cuda=cuda)
    return labels, costs, min_global_rotor


def REFORM_sequential(reference_model, query_model, n_samples=100, objects_per_sample=5,
                      iterations=100, covergence_threshold=0.00000001,
                      pool_size=1, object_type='generic', cuda=False, start_labels=None):

    #  Get the starting labels
    if start_labels is None:
        labels, costs = assign_measurements_to_objects_matrix(reference_model, query_model,
                                                              cuda=cuda)
    else:
        labels = [+l for l in start_labels]
    min_global_cost = np.inf
    min_global_rotor = 1.0 + 0.0 * e1

    r_est = 1.0 + 0.0 * e1
    remapped_objects = [o for o in query_model]

    assert iterations > 0, 'Must have at least 1 iteration'
    for i in range(iterations):
        # Reorder and estimate the rotor
        reordered_list_a = [reference_model[i] for i in labels]
        r_est_update, cost = estimate_rotor_objects_subsample_sequential(reordered_list_a, remapped_objects,
                                                                         n_samples,
                                                                         objects_per_sample,
                                                                         pool_size=pool_size,
                                                                         object_type=object_type)
        r_est = (r_est_update * r_est)
        r_est = r_est.normal()
        # Re map with our new rotor
        remapped_objects = [apply_rotor(l, r_est).normal() for l in query_model]
        # Get the new matching
        labels, costs = assign_measurements_to_objects_matrix(reference_model, remapped_objects,
                                                              object_type=object_type, cuda=cuda)
        current_cost = np.sum(costs)
        print(i, current_cost, covergence_threshold)
        if current_cost < min_global_cost:
            min_global_cost = current_cost
            min_global_rotor = +r_est
        if current_cost < covergence_threshold:
            return labels, costs, r_est

    print('Finished iterations')
    # Re map with our new rotor
    remapped_objects = [apply_rotor(l, min_global_rotor).normal() for l in query_model]
    # Get the new matching
    print('Rematching')
    labels, costs = assign_measurements_to_objects_matrix(reference_model, remapped_objects, cuda=cuda)
    print('REFORM complete')
    return labels, costs, min_global_rotor


def REFORM_cuda(reference_model, query_model, n_samples=100, objects_per_sample=5, iterations=100,
                covergence_threshold=0.00000001, mutation_probability=None, start_labels=None):
    #  Get the starting labels
    if start_labels is None:
        labels, costs = assign_measurements_to_objects_matrix(reference_model, query_model,
                                                              cuda=True)
    else:
        labels = [+l for l in start_labels]
    min_global_cost = np.inf
    min_global_rotor = 1.0 + 0.0 * e1

    r_est = 1.0 + 0.0 * e1
    remapped_objects = [o for o in query_model]
    assert iterations > 0, 'Must have at least 1 iteration'
    for i in range(iterations):
        # Reorder and estimate the rotor
        reordered_list_a = [reference_model[i] for i in labels]
        r_list, cost_array = sequential_rotor_estimation_cuda_mvs(reordered_list_a,
                                                                  remapped_objects,
                                                                  n_samples,
                                                                  objects_per_sample,
                                                                  mutation_probability=mutation_probability)
        min_cost_index = np.argmin(cost_array)
        min_cost = cost_array[min_cost_index]
        r_est_update = r_list[min_cost_index]
        r_est = (r_est_update * r_est)
        r_est = r_est.normal()
        # Re map with our new rotor
        remapped_objects = [apply_rotor(l, r_est).normal() for l in query_model]
        # Get the new matching
        labels, costs = assign_measurements_to_objects_matrix(reference_model, remapped_objects, cuda=True)
        current_cost = np.sum(costs)
        print(i, covergence_threshold, current_cost, min_global_cost)
        if current_cost < min_global_cost:
            min_global_cost = current_cost
            min_global_rotor = +r_est
        if current_cost < covergence_threshold:
            return labels, costs, r_est
    # Re map with our new rotor
    remapped_objects = [apply_rotor(l, min_global_rotor).normal() for l in query_model]
    # Get the new matching
    labels, costs = assign_measurements_to_objects_matrix(reference_model, remapped_objects, cuda=True)
    return labels, costs, min_global_rotor
